package com.sillycycle.bagleyd.rubik;

/*
 * @(#)Rubik3D.java
 *
 * Copyright 1994 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;

/**
 * The <code>Rubik3D</code> class does all the 3D graphics for
 * <code>RubikFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Rubik3D extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int TR = 8;
	static final int BR = 9;
	static final int BL = 10;
	static final int TL = 11;
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int COORD3D = 12;
	static final int MAX_VIEWS = 2;

	static final double SQRT_3 = 1.732050808;

	static int multiply(int a) {
		return ((int) (a * SQRT_3 / 2.0));
	}

	static int divide(int a) {
		return ((int) (a * 2.0 / SQRT_3));
	}

	transient RubikNextRow[] rotateToRow = /*CW to min face */
		new RubikNextRow[Rubik.MAX_FACES];
	static final int[/*MAX_RECT*/] planeToCube =
		{6, 0, 6, 1, 2, 3, 6, 4, 6, 6, 5, 6};
	static int[/*Rubik.MAX_FACES*/] cubeToPlane =
		{1, 3, 4, 5, 7, 10};

	Point pos = new Point(0, 0);
	Point puzzleSize, coreSize, puzzleOffset;
	Point viewMiddle, faceSize, viewSize;
	Point cubeSizeX, cubeSizeY, cubeSizeZ;
	int cubeDiagonalX, cubeDiagonalY, cubeDiagonalZ;
	int faceDiagonal, delta;
	int lastX, lastY;

	Point[][] faceLoc3D = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT];
	Point[][] cubeLoc3D = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT + 1];
	Point[] letter3DList = new Point[Rubik.MAX_FACES];
	Point[][][] orient3DList = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT][2];
	Point start = new Point(0, 0);

	JFrame      frame = null;
	Rubik       rubik = null;
	int         randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false, vertical = false;

	public Rubik3D(JFrame frame, Rubik rubik) {
		this.frame = frame;
		this.rubik = rubik;
		firstPaint = true;
	}

	public boolean checkStarted() {
		return rubik.checkStarted();
	}

	void cubeOffset3D(int face, int x, int y) {
		if (vertical)
			switch (face) {
			case 0:
				start = new Point(
					viewMiddle.x + delta - 1 +
					y * (cubeSizeX.x + delta) -
					x * (cubeDiagonalZ + delta),
					viewMiddle.y - delta - 2 -
					x * (cubeSizeZ.y + delta));
				break;
			case 1:
				start = new Point(
					viewMiddle.x - 2 * delta -
					x * (cubeDiagonalY + delta) -
					y * (cubeDiagonalZ + delta),
					viewMiddle.y +
					x * (cubeSizeY.y + delta) -
					y * (cubeSizeZ.y + delta));
				break;
			case 2:
				start = new Point(
					viewMiddle.x + delta - 1 +
					x * (cubeSizeX.x + delta) -
					y * (cubeDiagonalY + delta),
					viewMiddle.y + 2 * delta - 1 +
					y * (cubeSizeY.y + delta));
				break;
			case 3:
				start = new Point(
					viewMiddle.x + 2 * delta +
					x * (cubeDiagonalZ + delta) +
					y * (cubeDiagonalY + delta),
					viewSize.y + viewMiddle.y +
					delta - 1 -
					x * (cubeSizeZ.y + delta) +
					y * (cubeSizeY.y + delta));
				break;
			case 4:
				start = new Point(
					viewMiddle.x - delta + 1 -
					x * (cubeSizeX.x + delta) +
					y * (cubeDiagonalZ + delta),
					viewSize.y + viewMiddle.y -
					delta + 1 -
					y * (cubeSizeZ.y + delta));
				break;
			case 5:
				start = new Point(
					viewMiddle.x - 2 -
					y * (cubeSizeX.x + delta) +
					x * (cubeDiagonalY + delta),
					viewSize.y + viewMiddle.y +
					2 * delta + 2 +
					x * (cubeSizeY.y + delta));
				break;
			default:
				System.err.println("cubeOffset3D: face " +
					face);
		} else
			switch (face) {
			case 0:
				start = new Point(
					viewMiddle.x +
					y * (cubeSizeX.x + delta) -
					x * (cubeSizeZ.x + delta),
					viewMiddle.y - 2 * delta + 1 -
					y * (cubeDiagonalX + delta) -
					x * (cubeDiagonalZ + delta));
				break;
			case 1:
				start = new Point(
					viewMiddle.x - delta - 2 -
					y * (cubeSizeZ.x + delta),
					viewMiddle.y + delta +
					x * (cubeSizeY.y + delta) -
					y * (cubeDiagonalZ + delta));
				break;
			case 2:
				start = new Point(
					viewMiddle.x + 2 * delta - 1 +
					x * (cubeSizeX.x + delta),
					viewMiddle.y + delta +
					y * (cubeSizeY.y + delta) -
					x * (cubeDiagonalX + delta));
				break;
			case 3:
				start = new Point(
					viewSize.x + viewMiddle.x -
					delta + 1 -
					x * (cubeSizeZ.x + delta),
					viewMiddle.y - delta -
					y * (cubeSizeY.y + delta) +
					x * (cubeDiagonalZ + delta));
				break;
			case 4:
				start = new Point(
					viewSize.x + viewMiddle.x +
					delta - 1 -
					y * (cubeSizeZ.x + delta) +
					x * (cubeSizeX.x + delta),
					viewMiddle.y + 2 * delta +
					y * (cubeDiagonalZ + delta) +
					x * (cubeDiagonalX + delta));
				break;
			case 5:
				start = new Point(
					viewSize.x + viewMiddle.x +
					2 * delta + 2 +
					y * (cubeSizeX.x + delta),
					viewMiddle.y - delta -
					x * (cubeSizeY.y + delta) +
					y * (cubeDiagonalX + delta));
				break;
			default:
				System.err.println("cubeOffset3D: face " +
					face);
			}
	}

	// sets start
	void mapTo3D(int face, int i, int j) {
		switch (face) {
		case 0:
			start = new Point(rubik.sizeZ - 1 - j, i);
			break;
		case 1:
			start = new Point(j, rubik.sizeZ - 1 - i);
			break;
		case 2:
			start = new Point(i, j);
			break;
		case 3:
			start = new Point(rubik.sizeZ - 1 - i,
				rubik.sizeY - 1 - j);
			break;
		case 4:
			start = new Point(rubik.sizeX - 1 - i,
				rubik.sizeZ - 1 - j);
			break;
		case 5:
			start = new Point(j, rubik.sizeX - 1 - i);
				break;
		default:
			System.err.println("mapTo3D: face " + face);
		}
	}

	/* not used */
	void mapFrom3D(int face, int x, int y) {
		switch (face) {
		case 0:
			start = new Point(y, rubik.sizeZ - 1 - x);
			break;
		case 1:
			start = new Point(rubik.sizeZ - 1 - y, x);
			break;
		case 2:
			start = new Point(x, y);
			break;
		case 3:
			start = new Point(rubik.sizeZ - 1 - x,
				rubik.sizeY - 1 - y);
			break;
		case 4:
			start = new Point(rubik.sizeX - 1 - x,
				rubik.sizeZ - 1 - y);
			break;
		case 5:
			start = new Point(rubik.sizeX - 1 - y, x);
			break;
		default:
			System.err.println("mapTo3D: face " + face);
		}
	}

	static int mapOrientFrom3D(int face, int corner) {
		switch (face) {
		case 0:
			return (corner + 2) % RubikInterface.MAX_ORIENT;
		case 1:
			return corner;
		case 2:
			return (corner + 3) % RubikInterface.MAX_ORIENT;
		case 3:
			return (corner + 1) % RubikInterface.MAX_ORIENT;
		case 4:
			return (corner + 1) % RubikInterface.MAX_ORIENT;
		case 5:
			return corner;
		default:
			System.err.println("mapFrom3D: face " + face);
			return -1;
		}
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}

	static void drawPolygon(Graphics g,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawSquare(int face, int position, int offsetPosition) {
		Graphics g = rubik.getGraphics();

		if (g == null)
			return;
		try {
			int rotation, i, j, l;
			int sizeOfRow = rubik.sizeRow(face);
			Color fillColor, lineColor;

			i = position % sizeOfRow;
			j = position / sizeOfRow;
			mapTo3D(face, i, j); // sets start
			cubeOffset3D(face, start.x, start.y); // sets start again
			cubeLoc3D[face][0].x = start.x;
			cubeLoc3D[face][0].y = start.y;
			rotation = rubik.cubeLoc[face][position].rotation;
			l = rubik.cubeLoc[face][position].face;
			if (offsetPosition == 0) {
				if (rubik.mono) {
					if (rubik.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = rubik.faceColor[l];
					lineColor = rubik.borderColor;
				}
			} else {
				if (rubik.mono) {
					if (rubik.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = rubik.faceColor[l];
					fillColor = rubik.borderColor;
				}
			}
			int[] tempListX = new int[RubikInterface.MAX_ORIENT];
			int[] tempListY = new int[RubikInterface.MAX_ORIENT];

			for (int k = 0; k < RubikInterface.MAX_ORIENT; k++) {
				tempListX[k] =
					cubeLoc3D[face][k].x +
					((k > 0) ?
					tempListX[k - 1] : 0);
				tempListY[k] =
					cubeLoc3D[face][k].y +
					((k > 0) ?
					tempListY[k - 1] : 0);
			}
			fillPolygon(g, fillColor, lineColor,
				0, 0,
				tempListX, tempListY,
				RubikInterface.MAX_ORIENT);
			if (rubik.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = (cubeDiagonalX + cubeDiagonalY + cubeDiagonalZ) / 4;

				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(rubik.faceNames.charAt(l));
				g.drawString(buf,
					start.x + letter3DList[face].x,
					start.y + letter3DList[face].y);
			}
			if (rubik.orientChoice) {
				g.setColor(lineColor);
				g.drawLine(start.x +
					orient3DList[face][rotation][0].x,
					start.y +
					orient3DList[face][rotation][0].y,
					start.x +
					orient3DList[face][rotation][1].x,
					start.y +
					orient3DList[face][rotation][1].y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < rubik.sizeFace(face); position++)
			drawSquare(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < Rubik.MAX_FACES; face++)
			drawFace(face);
	}

	void drawFrame(boolean show) {
		Graphics g = rubik.getGraphics();

		if (g == null)
			return;
		try {
			if (show) {
				Point coord = new Point(viewSize.x + puzzleOffset.x,
					viewSize.y + puzzleOffset.y);
				int fontSize = faceDiagonal / 3;

				g.setFont(new Font("arial", 0, fontSize));
				g.setColor(rubik.foreground);
				if (vertical) {
					g.drawLine(0, coord.y,
						coord.x + puzzleOffset.x + 1, coord.y);
					g.drawString("Front", 2 * delta,
						3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
						-3 * delta - 10 * fontSize / 4 + coreSize.x,
						-delta - 2 * fontSize / 5 + coreSize.y);
				} else {
					g.drawLine(coord.x, 0,
						coord.x, coord.y + puzzleOffset.y + 1);
					g.drawString("Front", 2 * delta,
						3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
						-3 * delta - 10 * fontSize / 4 + coreSize.x,
						-delta - 2 * fontSize / 5 + coreSize.y);
				}
				g.setColor(rubik.currentForeground);
				for (int face = 0; face < Rubik.MAX_FACES; face++) {
					int[] tempListX = new int[RubikInterface.MAX_ORIENT];
					int[] tempListY = new int[RubikInterface.MAX_ORIENT];

					for (int i = 0; i < RubikInterface.MAX_ORIENT; i++) {
						tempListX[i] =
							faceLoc3D[face][i].x +
							((i > 0) ?
							tempListX[i - 1] : 0);
						tempListY[i] =
							faceLoc3D[face][i].y +
							((i > 0) ?
							tempListY[i - 1] : 0);
					}
					drawPolygon(g, 0, 0,
						tempListX, tempListY,
						RubikInterface.MAX_ORIENT);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(rubik.background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void resizePieces() {
		int face, orient, side, corner, sizex, sizey;
		Point[][] subcubeLoc3D = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT];
		Point[][] diamondLoc3D = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT];
		Point[][] subdiamondLoc3D = new Point[Rubik.MAX_FACES][RubikInterface.MAX_ORIENT];

		viewMiddle = new Point(faceSize.x + puzzleOffset.x,
			faceSize.y + puzzleOffset.y);
		for (face = 0; face < Rubik.MAX_FACES; face++) {
			faceLoc3D[face][0] = new Point(viewMiddle.x,
				viewMiddle.y);
			for (orient = 1; orient < RubikInterface.MAX_ORIENT; orient++) {
				faceLoc3D[face][orient] = new Point(faceSize.x,
					faceSize.y);
			}
		}
		if (vertical) {
			faceLoc3D[0][1].x /= -2;
			faceLoc3D[0][1].y /= -1;
			faceLoc3D[0][2].y = 0;
			faceLoc3D[0][3].x /= 2;

			faceLoc3D[1][1].x /= -2;
			faceLoc3D[1][2].x /= -2;
			faceLoc3D[1][2].y /= -1;
			faceLoc3D[1][3].x /= 2;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y = 0;
			faceLoc3D[2][2].x /= -2;
			faceLoc3D[2][3].x /= -1;
			faceLoc3D[2][3].y = 0;

			for (face = Rubik.MAX_FACES / 2; face < Rubik.MAX_FACES; face++)
				faceLoc3D[face][0].y += viewSize.y + 4;

			faceLoc3D[3][1].x /= 2;
			faceLoc3D[3][1].y /= -1;
			faceLoc3D[3][2].x /= 2;
			faceLoc3D[3][3].x /= -2;

			faceLoc3D[4][1].x /= -1;
			faceLoc3D[4][1].y = 0;
			faceLoc3D[4][2].x /= 2;
			faceLoc3D[4][2].y /= -1;
			faceLoc3D[4][3].y = 0;

			faceLoc3D[5][1].x /= 2;
			faceLoc3D[5][2].x /= -1;
			faceLoc3D[5][2].y = 0;
			faceLoc3D[5][3].x /= -2;
			faceLoc3D[5][3].y /= -1;
		} else {
			faceLoc3D[0][1].x /= -1;
			faceLoc3D[0][1].y /= -2;
			faceLoc3D[0][2].y /= -2;
			faceLoc3D[0][3].y /= 2;

			faceLoc3D[1][1].x = 0;
			faceLoc3D[1][2].x /= -1;
			faceLoc3D[1][2].y /= -2;
			faceLoc3D[1][3].x = 0;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y /= -2;
			faceLoc3D[2][2].x = 0;
			faceLoc3D[2][3].y /= 2;
			faceLoc3D[2][3].x /= -1;

			for (face = Rubik.MAX_FACES / 2; face < Rubik.MAX_FACES; face++)
				faceLoc3D[face][0].x += viewSize.x + 4;

			faceLoc3D[3][1].x /= -1;
			faceLoc3D[3][1].y /= 2;
			faceLoc3D[3][2].x = 0;
			faceLoc3D[3][2].y /= -1;
			faceLoc3D[3][3].y /= -2;

			faceLoc3D[4][1].y /= 2;
			faceLoc3D[4][2].x /= -1;
			faceLoc3D[4][2].y /= 2;
			faceLoc3D[4][3].x /= -1;
			faceLoc3D[4][3].y /= -2;

			faceLoc3D[5][1].x = 0;
			faceLoc3D[5][1].y /= -1;
			faceLoc3D[5][2].y /= 2;
			faceLoc3D[5][3].x = 0;
		}
		for (face = 0; face < Rubik.MAX_FACES; face++) {
			cubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			subcubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			switch (face) {
			case 0:
			case 4:
				sizex = rubik.sizeX;
				sizey = rubik.sizeZ;
				break;
			case 1:
			case 3:
				sizex = rubik.sizeZ;
				sizey = rubik.sizeY;
				break;
			default:
				/* if (face == 2 || face == 5) */
				sizex = rubik.sizeX;
				sizey = rubik.sizeY;
				break;
			}
			for (orient = 1; orient < RubikInterface.MAX_ORIENT; orient++) {
				cubeLoc3D[face][orient] = new Point(0, 0);
				if ((((orient & 1) == 1) && (face == 2 || face == 3 || face == 4)) ||
					(((orient & 1) == 0) && !(face == 2 || face == 3 || face == 4)))
					cubeLoc3D[face][orient].x =
						(faceLoc3D[face][orient].x - (3 + sizex) *
					 delta * faceLoc3D[face][orient].x /
						 faceSize.x) / sizex;
				else
					cubeLoc3D[face][orient].x =
						(faceLoc3D[face][orient].x - (3 + sizey) *
					 delta * faceLoc3D[face][orient].x /
						 faceSize.x) / sizey;
				if ((((orient & 1) == 1) && !(face == 2 || face == 3 || face == 4)) ||
					(((orient & 1) == 0) && (face == 2 || face == 3 || face == 4)))
					cubeLoc3D[face][orient].y =
						(faceLoc3D[face][orient].y - (3 + sizey) *
					 delta * faceLoc3D[face][orient].y /
						 faceSize.y) / sizey;
				else
					cubeLoc3D[face][orient].y =
						(faceLoc3D[face][orient].y - (3 + sizex) *
					 delta * faceLoc3D[face][orient].y /
						 faceSize.y) / sizex;
				subcubeLoc3D[face][orient] = new Point(0, 0);
				if ((((orient & 1) == 1) && (face == 2 || face == 3 || face == 4)) ||
					(((orient & 1) == 0) && !(face == 2 || face == 3 || face == 4)))
					subcubeLoc3D[face][orient].x =
						(faceLoc3D[face][orient].x - (5 + sizex) *
					 delta * faceLoc3D[face][orient].x /
						 faceSize.x) / (2 * sizex);
				else
					subcubeLoc3D[face][orient].x =
						(faceLoc3D[face][orient].x - (5 + sizey) *
					 delta * faceLoc3D[face][orient].x /
						 faceSize.x) / (2 * sizey);
				if ((((orient & 1) == 1) && !(face == 2 || face == 3 || face == 4)) ||
					(((orient & 1) == 0) && (face == 2 || face == 3 || face == 4)))
					subcubeLoc3D[face][orient].y =
						(faceLoc3D[face][orient].y - (5 + sizey) *
					 delta * faceLoc3D[face][orient].y /
						 faceSize.y) / (2 * sizey);
				else
					subcubeLoc3D[face][orient].y =
						(faceLoc3D[face][orient].y - (5 + sizex) *
					 delta * faceLoc3D[face][orient].y /
						 faceSize.y) / (2 * sizex);
			}
			cubeLoc3D[face][RubikInterface.MAX_ORIENT] = new Point(-cubeLoc3D[face][1].x -
				cubeLoc3D[face][2].x - cubeLoc3D[face][3].x,
				-cubeLoc3D[face][1].y -
				cubeLoc3D[face][2].y - cubeLoc3D[face][3].y);
		}
		cubeSizeX = new Point(
			java.lang.Math.max(faceSize.x / rubik.sizeX - delta, 0),
			java.lang.Math.max(faceSize.y / rubik.sizeX - delta, 0));
		cubeSizeY = new Point(
			java.lang.Math.max(faceSize.x / rubik.sizeY - delta, 0),
			java.lang.Math.max(faceSize.y / rubik.sizeY - delta, 0));
		cubeSizeZ = new Point(
			java.lang.Math.max(faceSize.x / rubik.sizeZ - delta, 0),
			java.lang.Math.max(faceSize.y / rubik.sizeZ - delta, 0));
		cubeDiagonalX = java.lang.Math.max((faceDiagonal - delta) /
			rubik.sizeX - delta, 1);
		cubeDiagonalY = java.lang.Math.max((faceDiagonal - delta) /
			rubik.sizeY - delta, 1);
		cubeDiagonalZ = java.lang.Math.max((faceDiagonal - delta) /
			rubik.sizeZ - delta, 1);
		if (vertical) {
			letter3DList[0] = new Point(
				cubeSizeX.x / 4, -cubeSizeZ.y / 4);
			letter3DList[1] = new Point(
				-4 * cubeDiagonalZ / 3, delta / 2);
			letter3DList[2] = new Point(
				cubeSizeX.x / 4, cubeSizeY.y / 2);
			letter3DList[3] = new Point(
				cubeDiagonalZ, delta / 2);
			letter3DList[4] = new Point(
				-cubeSizeX.x / 4, -cubeSizeZ.y / 4);
			letter3DList[5] = new Point(
				-cubeSizeX.x / 4, cubeSizeY.y / 2);
		} else {
			letter3DList[0] = new Point(
				-delta, -cubeDiagonalZ);
			letter3DList[1] = new Point(
				-cubeSizeZ.x / 2, cubeSizeY.y / 4);
			letter3DList[2] = new Point(
				cubeSizeX.x / 4, cubeSizeY.y / 4);
			letter3DList[3] = new Point(
				-cubeSizeZ.x / 2, -cubeSizeY.y / 6);
			letter3DList[4] = new Point(
				-delta, 3 * cubeDiagonalZ / 2);
			letter3DList[5] = new Point(
				cubeSizeX.x / 4, -cubeSizeY.y / 8);
		}
		/* The following figures out where to put the orient lines */
		for (face = 0; face < Rubik.MAX_FACES; face++) {
			for (orient = 0; orient < RubikInterface.MAX_ORIENT - 1; orient++) {
				diamondLoc3D[face][orient] = new Point(
					(cubeLoc3D[face][orient].x +
					cubeLoc3D[face][orient + 1].x) / 2,
					(cubeLoc3D[face][orient].y +
					cubeLoc3D[face][orient + 1].y) / 2);
				subdiamondLoc3D[face][orient] = new Point(
					(subcubeLoc3D[face][orient].x +
					subcubeLoc3D[face][orient + 1].x) / 2,
					(subcubeLoc3D[face][orient].y +
					subcubeLoc3D[face][orient + 1].y) / 2);
			}
			/* Its a parallelagram so take advantage of that */
			diamondLoc3D[face][orient] = new Point(
				(cubeLoc3D[face][RubikInterface.MAX_ORIENT - 1].x -
				cubeLoc3D[face][RubikInterface.MAX_ORIENT / 2].x) / 2,
				(cubeLoc3D[face][RubikInterface.MAX_ORIENT - 1].y -
				cubeLoc3D[face][RubikInterface.MAX_ORIENT / 2].y) / 2);
			subdiamondLoc3D[face][orient] = new Point(
				(subcubeLoc3D[face][RubikInterface.MAX_ORIENT - 1].x -
				subcubeLoc3D[face][RubikInterface.MAX_ORIENT / 2].x) / 2,
				(subcubeLoc3D[face][RubikInterface.MAX_ORIENT - 1].y -
				subcubeLoc3D[face][RubikInterface.MAX_ORIENT / 2].y) / 2);

			corner = mapOrientFrom3D(face, 1);
			orient3DList[face][corner][0] = new Point(
				cubeLoc3D[face][1].x / 2,
				cubeLoc3D[face][1].y / 2);
			orient3DList[face][corner][1] = new Point(
				orient3DList[face][corner][0].x +
				(cubeLoc3D[face][2].x - subcubeLoc3D[face][2].x) / 2,
				orient3DList[face][corner][0].y +
				(cubeLoc3D[face][2].y - subcubeLoc3D[face][2].y) / 2);
			for (orient = 1; orient < RubikInterface.MAX_ORIENT; orient++) {
				side = corner;
				corner = mapOrientFrom3D(face, (orient + 1) % RubikInterface.MAX_ORIENT);
				orient3DList[face][corner][0] = new Point(
					orient3DList[face][side][0].x +
					diamondLoc3D[face][orient].x,
					orient3DList[face][side][0].y +
					diamondLoc3D[face][orient].y);
				orient3DList[face][corner][1] = new Point(
					orient3DList[face][side][1].x +
					subdiamondLoc3D[face][orient].x,
					orient3DList[face][side][1].y +
					subdiamondLoc3D[face][orient].y);
			}
		}
	}

	void narrowSelection() {
		// sets rubik.currentFace, rubik.currentPosition, rubik.currentDirection
		switch (rubik.currentDirection) {
		case Rubik.TOP:
		case Rubik.RIGHT:
		case Rubik.BOTTOM:
		case Rubik.LEFT:
			if (vertical) {
				if (rubik.currentFace == 1 ||
						rubik.currentFace == 3) {
					rubik.currentFace = -1;
					return;
				}
			} else {
				if (rubik.currentFace == 0 ||
						rubik.currentFace == 4) {
					rubik.currentFace = -1;
					return;
				}
				if (rubik.currentFace == 5)
					rubik.currentDirection =
						(rubik.currentDirection + 2) %
						RubikInterface.MAX_ORIENT;
			}
			break;
		case Rubik.CCW:
		case Rubik.CW:
			break;
		case TR:
			if (vertical) {
				switch (rubik.currentFace) {
				case 0:
				case 5:
					rubik.currentFace = -1;
					return;
				case 1:
				case 2:
				case 4:
					rubik.currentDirection =
						Rubik.TOP;
					break;
				default:
					/* (rubik.currentFace == 3) */
					rubik.currentDirection = Rubik.LEFT;
					break;
				}
			} else {
				switch (rubik.currentFace) {
				case 1:
				case 5:
					rubik.currentFace = -1;
					return;
				case 0:
				case 2:
				case 3:
					rubik.currentDirection = Rubik.RIGHT;
					break;
				default:
					/* (rubik.currentFace == 4) */
					rubik.currentDirection = Rubik.BOTTOM;
					break;
				}
			}
			break;
		case BR:
			if (vertical) {
				switch (rubik.currentFace) {
				case 2:
				case 4:
					rubik.currentFace = -1;
					return;
				case 0:
				case 5:
					rubik.currentDirection = Rubik.BOTTOM;
					break;
				case 1:
					rubik.currentDirection = Rubik.RIGHT;
					break;
				default:
					/* (rubik.currentFace == 3) */
					rubik.currentDirection = Rubik.TOP;
					break;
				}
			} else {
				switch (rubik.currentFace) {
				case 2:
				case 3:
					rubik.currentFace = -1;
					return;
				case 4:
				case 5:
					rubik.currentDirection = Rubik.LEFT;
					break;
				case 0:
					rubik.currentDirection = Rubik.BOTTOM;
					break;
				default:
					/* (rubik.currentFace == 1) */
					rubik.currentDirection = Rubik.RIGHT;
					break;
				}
			}
			break;
		case BL:
			if (vertical) {
				switch (rubik.currentFace) {
				case 0:
				case 5:
					rubik.currentFace = -1;
					return;
				case 1:
				case 2:
				case 4:
					rubik.currentDirection = Rubik.BOTTOM;
					break;
				default:
					/* (rubik.currentFace == 3) */
					rubik.currentDirection = Rubik.RIGHT;
					break;
				}
			} else {
				switch (rubik.currentFace) {
				case 1:
				case 5:
					rubik.currentFace = -1;
					return;
				case 0:
				case 2:
				case 3:
					rubik.currentDirection = Rubik.LEFT;
					break;
				default:
					/* (rubik.currentFace == 4) */
					rubik.currentDirection = Rubik.TOP;
					break;
				}
			}
			break;
		case TL:
			if (vertical) {
				switch (rubik.currentFace) {
				case 2:
				case 4:
					rubik.currentFace = -1;
					return;
				case 0:
				case 5:
					rubik.currentDirection = Rubik.TOP;
					break;
				case 1:
					rubik.currentDirection = Rubik.LEFT;
					break;
				default:
					/* (rubik.currentFace == 3) */
					rubik.currentDirection = Rubik.BOTTOM;
					break;
				}
			} else {
				switch (rubik.currentFace) {
				case 2:
				case 3:
					rubik.currentFace = -1;
					return;
				case 4:
				case 5:
					rubik.currentDirection = Rubik.RIGHT;
					break;
				case 0:
					rubik.currentDirection = Rubik.TOP;
					break;
				default:
					/* (rubik.currentFace == 1) */
					rubik.currentDirection = Rubik.LEFT;
					break;
				}
			}
			break;
		default:
			rubik.currentFace = -1;
			return;
		}
		/* Remap to row movement */
		if (rubik.currentDirection == Rubik.CW ||
				rubik.currentDirection == Rubik.CCW) {
			int i, j, sizeOfRow, sizeOfColumn, newFace;

			newFace = rotateToRow[rubik.currentFace].face;
			sizeOfRow = rubik.sizeRow(newFace);
			sizeOfColumn = rubik.sizeColumn(newFace);
			rubik.currentDirection =
				(rubik.currentDirection == Rubik.CCW) ?
				(rotateToRow[rubik.currentFace].direction + 2) % RubikInterface.MAX_ORIENT :
				rotateToRow[rubik.currentFace].direction;
			if (rotateToRow[rubik.currentFace].sideFace == Rubik.LEFT ||
				rotateToRow[rubik.currentFace].sideFace == Rubik.BOTTOM) {
				i = sizeOfRow - 1;
				j = sizeOfColumn - 1;
			} else {
				i = j = 0;
			}
			rubik.currentFace = newFace;
			rubik.currentPosition = j * sizeOfRow + i;
		}
	}

	int positionToPiece(int x, int y) { // sets rubik.currentPosition
		int u, v, i, j, face;
		int tempx = x, tempy = y, sizeOfRow, sizeOfColumn;
		boolean front, tl, ur, ul;

		if (vertical) {
			tempx -= viewMiddle.x;
			front = (tempy < viewSize.y + puzzleOffset.y);
			if (!front)
				tempy -= (viewSize.y);
			tl = (tempy < viewMiddle.y);
			tempy -= viewMiddle.y;
			u = -faceSize.y * tempx + faceDiagonal * tempy;
			v = faceSize.y * tempx + faceDiagonal * tempy;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 0 : 1);
				else
					face = ((ul) ? 1 : 2);
			} else {
				if (tl)
					face = ((ul) ? 4 : 3);
				else
					face = ((ur) ? 3 : 5);
			}
		} else {
			tempy -= viewMiddle.y;
			front = (tempx < viewSize.x + puzzleOffset.x);
			if (!front)
				tempx -= (viewSize.x);
			tl = (tempx < viewMiddle.x);
			tempx -= viewMiddle.x;
			u = -faceSize.x * tempy + faceDiagonal * tempx;
			v = faceSize.x * tempy + faceDiagonal * tempx;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 1 : 0);
				else
					face = ((ul) ? 0 : 2);
			} else {
				if (tl)
					face = ((ul) ? 3 : 4);
				else
					face = ((ur) ? 4 : 5);
			}
		}
		sizeOfRow = rubik.sizeRow(face);
		sizeOfColumn = rubik.sizeColumn(face);
		if (vertical)
			switch (face) {
			case 0:
				i = (tempx - 2 -
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSizeX.x + delta);
				j = (tempy + 2) / (cubeSizeZ.y + delta) +
					sizeOfColumn - 1;
				break;
			case 1:
				i = (tempx + 4 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonalZ + delta)) + sizeOfRow - 1;
				j = (-tempx - 6 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonalY + delta));
				break;
			case 2:
				i = (tempx - 4 +
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSizeX.x + delta);
				j = (tempy - 4) / (cubeSizeY.y + delta);
				break;
			case 3:
				i = (-tempx + 5 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonalZ + delta)) + sizeOfRow - 1;
				j = (-tempx + 7 -
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonalY + delta)) + sizeOfColumn - 1;
				break;
			case 4:
				i = (tempx +
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSizeX.x + delta) + sizeOfRow - 1;
				j = (tempy - 2) / (cubeSizeZ.y + delta) +
					sizeOfColumn - 1;
				break;
			case 5:
				i = (tempx + 2 -
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSizeX.x + delta) + sizeOfRow - 1;
				j = (tempy - 6) / (cubeSizeY.y + delta);
				break;
			default:
				face = -1;
				return face;
		} else
			switch (face) {
			case 0:
				i = (-tempy - 3 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonalX + delta));
				j = (tempy + 1 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonalZ + delta)) + sizeOfColumn - 1;
				break;
			case 1:
				i = (tempx + 2) / (cubeSizeZ.x + delta) +
					sizeOfRow - 1;
				j = (tempy - 3 -
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSizeY.y + delta);
				break;
			case 2:
				i = (tempx - 4) / (cubeSizeX.x + delta);
				j = (tempy - 6 +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSizeY.y + delta);
				break;
			case 3:
				i = tempx / (cubeSizeZ.x + delta) +
					sizeOfRow - 1;
				j = (tempy +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSizeY.y + delta) + sizeOfColumn - 1;
				break;
			case 4:
				i = (-tempy + 9 -
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonalX + delta)) + sizeOfRow - 1;
				j = (-tempy + 7 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonalZ + delta)) + sizeOfColumn - 1;
				break;
			case 5:
				i = (-tempx + 7) / (cubeSizeX.x + delta) +
					sizeOfRow - 1;
				j = (-tempy - 4 +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSizeY.y + delta);
				break;
			default:
				face = -1;
				return face;
			}
		if (i < 0 || j < 0 || i >= sizeOfRow || j >= sizeOfColumn) {
			face = -1;
			return face;
		}
		rubik.currentPosition = j * sizeOfRow + i;
		return face;
	}

	void resizePuzzle() {
		coreSize = new Point(rubik.getSize().width,
			rubik.getSize().height);
		delta = 4;
		vertical = (coreSize.y >= coreSize.x);
		/* Unlike the 2d implementation... lets have variable size
			 cubies and static sized cube */
		if (vertical) {
			viewSize = new Point(coreSize.x,
				coreSize.y / MAX_VIEWS);
			if (viewSize.x >= divide(viewSize.y)) {
				viewSize.y = java.lang.Math.max(viewSize.y, 0);
				viewSize.x = divide(viewSize.y);
			} else {
				viewSize.x = java.lang.Math.max(viewSize.x, 0);
				viewSize.y = multiply(viewSize.x);
			}
			faceSize = new Point(
				java.lang.Math.max(viewSize.x / 2 - 2, 1),
				java.lang.Math.max(viewSize.y / 2 - 2, 1));
			faceDiagonal = faceSize.x / 2;
			puzzleSize = new Point(viewSize.x,
				MAX_VIEWS * viewSize.y);
		} else {
			viewSize = new Point(coreSize.x / MAX_VIEWS,
				coreSize.y);
			if (viewSize.y >= divide(viewSize.x)) {
				viewSize.x = java.lang.Math.max(viewSize.x, 0);
				viewSize.y = divide(viewSize.x);
			} else {
				viewSize.y = java.lang.Math.max(viewSize.y, 0);
				viewSize.x = multiply(viewSize.y);
			}
			faceSize = new Point(
				java.lang.Math.max(viewSize.x / 2 - 2, 1),
				java.lang.Math.max(viewSize.y / 2 - 2, 1));
			faceDiagonal = faceSize.y / 2;
			puzzleSize = new Point(MAX_VIEWS * viewSize.x,
				viewSize.y);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		rotateToRow[0] = new RubikNextRow(1, Rubik.LEFT, Rubik.TOP);
		rotateToRow[1] = new RubikNextRow(0, Rubik.BOTTOM, Rubik.RIGHT);
		rotateToRow[2] = new RubikNextRow(0, Rubik.RIGHT, Rubik.BOTTOM);
		rotateToRow[3] = new RubikNextRow(0, Rubik.TOP, Rubik.LEFT);
		rotateToRow[4] = new RubikNextRow(1, Rubik.RIGHT, Rubik.BOTTOM);
		rotateToRow[5] = new RubikNextRow(0, Rubik.LEFT, Rubik.TOP);
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		rubik.drawAllPieces();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if ((rubik.currentFace = positionToPiece(x, y)) != -1) { //sets currentPosition
			if (control || rubik.practiceChoice || !rubik.checkSolved(rubik.orientChoice)) {
				selectPaint = true;
				rubik.repaint();
				return;
			}
			rubik.currentFace = -1;
		}
		rubik.currentFace = -1;
	}

	void releasePuzzle(boolean control, int x, int y) {
		if (rubik.currentFace == -1)
			return;
		rubik.nextControl = control;
		int tempPosition = rubik.currentPosition;
		rubik.nextFace = positionToPiece(x, y); // sets currentPosition
		rubik.nextPosition = rubik.currentPosition;
		rubik.currentPosition = tempPosition;
		releasePaint = true;
		rubik.repaint();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			rubik.initializePuzzle();
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (rubik.currentFace != -1) {
				drawSquare(rubik.currentFace,
					rubik.currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (rubik.currentFace != -1) {
				drawSquare(rubik.currentFace,
					rubik.currentPosition, 0);
				rubik.selectPieces();
				rubik.currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (rubik.currentControl)
				rubik.moveControlCb(rubik.currentFace,
					rubik.currentDirection);
			else {
				rubik.movePieces(rubik.currentFace,
					rubik.currentPosition,
					rubik.currentDirection);
				if (stackPaint && rubik.checkSolved(rubik.orientChoice)) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_SOLVED,
						3);
				}
			}
			rubik.currentDirection = rubik.currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(rubik.delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			rubik.randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public void keyPressed(KeyEvent event) {
		int modifiers = event.getModifiersEx();

		switch (event.getKeyCode()) {
		case KeyEvent.VK_F11:
		case KeyEvent.VK_DIVIDE:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.CCW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_HOME:
		case KeyEvent.VK_NUMPAD7:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				TL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.TOP,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_UP:
		case KeyEvent.VK_NUMPAD9:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				TR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.LEFT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_F12:
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.CW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.RIGHT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_END:
		case KeyEvent.VK_NUMPAD1:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				BL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.BOTTOM,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_DOWN:
		case KeyEvent.VK_NUMPAD3:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				BR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		default:
		}
	}
}
