package com.sillycycle.bagleyd.rubik;

/*
 * @(#)RubikFrame.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.Colour;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>RubikFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class RubikFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_CUBES = 7;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "rubik.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Practice", "Randomize", "Auto-solve",
		"Increment", "Decrement", "Orientize",
		"Speed >", "Slow <", "Sound @",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'p', 'z', 's',
		'i', 'd', 'o', '>', '<', '@',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 320;
	int heightValue = 175;
	static int xValue = 0;
	static int yValue = 0;
	String monoString = "RYWGOB";
	Color[] faceColor = new Color[] {
		Color.red, Color.yellow, Color.white, Color.green,
		Color.orange, Color.blue,
	};
	Color borderColor = Color.black;
	Color fgColor = Color.cyan;
	Color bgColor = Colour.steelBlue;
	boolean controlValue = true;
	boolean sound = false;
	int sizeXValue = RubikInterface.DEFAULTCUBES;
	int sizeYValue = RubikInterface.DEFAULTCUBES;
	int sizeZValue = RubikInterface.DEFAULTCUBES;
	boolean orientValue = RubikInterface.DEFAULT_ORIENT;
	boolean practiceValue = RubikInterface.DEFAULT_PRACTICE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean debug = false;
	JPanel topPanel = new JPanel();
	JCheckBox orientCheckBox = new JCheckBox("Oriented");
	JCheckBox practiceCheckBox = new JCheckBox("Practice");
	RubikStack undo = new RubikStack();
	RubikStack redo = new RubikStack();
	Rubik puzzle2D, puzzle3D;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Rubik! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][][][] puzzleRecord = null;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner sizeXChanger = null, sizeYChanger = null, sizeZChanger = null;
	Spinner speedChanger = null;
	//Slider sizeXChanger = null, sizeYChanger = null, sizeZChanger = null;
	//Slider speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle2D.stringSave = s;
		puzzle3D.stringSave = s;
	}

	public void getPuzzle() {
		puzzle2D.getPuzzle(false);
		puzzle3D.getPuzzle(true);
	}

	public void getPuzzleFile() {
		puzzle2D.getPuzzleFile(false);
		puzzle3D.getPuzzleFile(true);
	}

	public void writePuzzle() {
		puzzle3D.writePuzzle();
	}

	public void undoPuzzle() {
		puzzle3D.undoPuzzle();
	}

	public void redoPuzzle() {
		puzzle3D.redoPuzzle();
	}

	public void clearPuzzle() {
		puzzle3D.clearPuzzle();
	}

	public void practicePuzzle() {
		puzzle3D.practicePuzzle();
	}

	public void randomizePuzzle() {
		puzzle3D.randomizePuzzle();
	}

	public void solvePuzzle() {
		puzzle3D.solvePuzzle();
	}

	public void orientizePuzzle() {
		puzzle3D.orientizePuzzle();
	}

	public void incrementPuzzle() {
		puzzle3D.incrementPuzzle();
	}

	public void decrementPuzzle() {
		puzzle3D.decrementPuzzle();
	}

	public void processKey(char key) {
		puzzle3D.processKey(key);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}


	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		for (int i = 0; i < faceColor.length; i++)
			faceColor[i] = getOpt("faceColor" + i, faceColor[i]);
		borderColor = getOpt("borderColor", borderColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		sizeXValue = getOpt("sizex", sizeXValue);
		if (sizeXValue < RubikInterface.MIN_CUBES)
			sizeXValue = RubikInterface.DEFAULTCUBES;
		sizeYValue = getOpt("sizey", sizeYValue);
		if (sizeYValue < RubikInterface.MIN_CUBES)
			sizeYValue = RubikInterface.DEFAULTCUBES;
		sizeZValue = getOpt("sizez", sizeZValue);
		if (sizeZValue < RubikInterface.MIN_CUBES)
			sizeZValue = RubikInterface.DEFAULTCUBES;
		orientValue = getOpt("orient", orientValue);
		practiceValue = getOpt("practice", practiceValue);
		delayValue = getOpt("delay", delayValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Rubik", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Rubik - the original puzzle has 9 squares per face (size = 3)." + NL +
"puzzle was designed by Erno Rubik and called the Rubik's Cube." + NL +
"This has 8!*12!*3^8*2^12/12 or 4.3 * 10^19 different" + NL +
"combinations." + NL +
"The Pocket Cube has 4 squares per face (size = 2) also" + NL +
"designed by Erno Rubik.  This has 7!*3^6 or 3,674,160" + NL +
"different combinations." + NL +
"Rubik's Revenge has 16 squares per face (size = 4) also" + NL +
"designed by Erno Rubik.  This has 7!*3^6*24!*24!/(4!)^6 or" + NL +
"7.4 * 10^46 different combinations." + NL +
"5x5x5 Cube.  This has 8!*12!*3^7*2^10*(24!)^3/(4!)^12 or" + NL +
"2.83 * 10^74 different combinations." + NL +
"There is also the Magic Domino 3x3x2 cube which has (8!)^2/4" + NL +
"or 406,425,600 combinations." + NL +
"A physical 6x6x6 cube is possible but to my knowledge no one" + NL +
"has been too successful in building one.  7x7x7 is also" + NL +
"possible, but here one must make the center most cubes smaller" + NL +
"than the outside cubes, so the corners do not fall off when" + NL +
"turned." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Make sure you randomize the puzzle first before trying to move the" + NL +
"pieces for the first time." + NL +
" " + NL +
"Press \"mouse-left\" button to move a piece.  Release \"mouse-left\"" + NL +
"button on another piece on the same face and in the same row." + NL +
"The pieces will then turn towards where the mouse button was" + NL +
"released." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"Z\" or \"z\" keys, to" + NL +
"randomize the puzzle (this must be done first to set a new record)." + NL +
"One must double click on \"mouse-right\" if the puzzle is currently" + NL +
"being worked on." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to buffer." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to buffer." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear puzzle." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to start auto-solver.  Only works on 1x1x1," + NL +
"2x2x2, 3x3x3, and 4x4x4 cubes." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the orient mode.  (Default is off)." + NL +
"One has to orient the faces in orient mode, besides getting all" + NL +
"the faces to be the same color.  To do this one has to get the" + NL +
"lines to be oriented in the same direction, this only matters with" + NL +
"center pieces, not on a corner or edge).  This does add" + NL +
"complexity so there are 2 sets of records." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to increase the number of pieces." + NL +
" " + NL +
"Press \"D\" or \"d\" keys to decrease the number of pieces." + NL +
" " + NL +
"Press \"P\" or \"p\" keys, to toggle the practice mode (in practice" + NL +
"mode the record should say \"practice\").  This is good for" + NL +
"learning moves and experimenting." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the pieces." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the pieces." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the mouse." + NL +
"Key pad (better with Num Lock on) is defined for Rubik2D as:" + NL +
"   /    Counterclockwise" + NL +
" " + NL +
"   8     Up" + NL +
"   ^" + NL +
"4<5>6  Left, Clockwise, Right" + NL +
"   v" + NL +
"   2     Down" + NL +
" " + NL +
"Key pad for Rubik3D, use must use your intuition (is this a cop" + NL +
"out or what?).  The key pad is defined differently depending on" + NL +
"which side of the cube your mouse is pointing at.  One thing that" + NL +
"stays the same is \"5\" is Clockwise and \"/\" is Counterclockwise." + NL +
" " + NL +
"Use the control key and the left mouse button, keypad, or arrow" + NL +
"keys to move the whole puzzle.  This is not recorded as a turn." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"Inside Rubik's Cube and Beyond by Christoph Bandelow, " +
"Birkhauser, 1982 pp 44, 45, 88, 89." + NL +
"Magic Cubes 1996 Catalog of Dr. Christoph Bandelow." + NL +
"The Simple Solution To Rubik's Cube, James G. Nourse, " +
"June 1981." + NL +
"Rubik's Cube Newsletter by Ideal Aug 1982 Vol.1 No. 2." + NL +
"Rubik's Cube The Solution, Ideal Toy Corporation, 1981." + NL +
"Rubik's Revenge Puzzle The Solution, Ideal Toy " +
"Corporation, 1982." + NL,

				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java RubikFrame Version 8.7.2" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = RubikFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = RubikFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playMoveAudio() {
		playClip(RubikInterface.SOUNDS[RubikInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle2D == null || puzzle3D == null)
			return;
		puzzle2D.setBounds(0, 0, width, height / 2);
		puzzle3D.setBounds(0, 0, width, height - height / 2);
		validate();
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int orient = 0; orient < puzzleRecord.length; orient++) {
			for (int i = 0; i < puzzleRecord[orient].length; i++) {
				for (int j = i; j < puzzleRecord[orient][i].length; j++) {
					for (int k = j; k < puzzleRecord[orient][i][j].length; k++) {
						scoreString.append(puzzleRecord[orient][i][j][k]);
						if (k < puzzleRecord[orient][i][j].length - 1)
							scoreString.append(" ");
					}
					if (j < puzzleRecord[orient][i].length - 1
							|| i < puzzleRecord[orient].length - 1
							|| orient < puzzleRecord.length - 1)
						scoreString.append(NL);
				}
				if (i < puzzleRecord[orient].length - 1
						|| orient < puzzleRecord.length - 1)
				scoreString.append(NL);
			}
			if (orient < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[2]
			[MAX_CUBES - RubikInterface.MIN_CUBES + 1]
			[MAX_CUBES - RubikInterface.MIN_CUBES + 1]
			[MAX_CUBES - RubikInterface.MIN_CUBES + 1];
		for (int orient = 0; orient < puzzleRecord.length; orient++) {
		  int z = 0;

		  for (int i = 0; i < puzzleRecord[orient].length; i++) {
		    for (int j = i; j < puzzleRecord[orient][i].length; j++) {
		      for (int k = j; k < puzzleRecord[orient][i][j].length; k++) {
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[orient][k][j][i] =
					puzzleRecord[orient][k][i][j] =
					puzzleRecord[orient][j][k][i] =
					puzzleRecord[orient][j][i][k] =
					puzzleRecord[orient][i][k][j] =
					puzzleRecord[orient][i][j][k] =
					RubikScores.PUZZLE_SCORES[orient][z++];
			} else {
				puzzleRecord[orient][k][j][i] =
					puzzleRecord[orient][k][i][j] =
					puzzleRecord[orient][j][k][i] =
					puzzleRecord[orient][j][i][k] =
					puzzleRecord[orient][i][k][j] =
					puzzleRecord[orient][i][j][k] =
					Integer.parseInt(st[field++]);
			}
		      }
		    }
		  }
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(int sizeX, int sizeY,
			int sizeZ, boolean orient,
			boolean practice) {
		int i = sizeX - RubikInterface.MIN_CUBES;
		int j = sizeY - RubikInterface.MIN_CUBES;
		int k = sizeZ - RubikInterface.MIN_CUBES;
		int l = (orient) ? 1 : 0;

		if (puzzleRecord == null)
			return;
		if (practice) {
			recordValueLabel.setText("PRACTICE");
		} else if (sizeX > MAX_CUBES || sizeY > MAX_CUBES ||
				sizeZ > MAX_CUBES) {
			recordValueLabel.setText("NONE");
		} else if (puzzleRecord[l][i][j][k] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[l][i][j][k]));
		}
	}

	boolean handleSolved(int counter, int sizeX, int sizeY,
			int sizeZ, boolean orient,
			boolean practice) {
		int i = sizeX - RubikInterface.MIN_CUBES;
		int j = sizeY - RubikInterface.MIN_CUBES;
		int k = sizeZ - RubikInterface.MIN_CUBES;
		int l = (orient) ? 1 : 0;
		int notL = (orient) ? 0 : 1;

		if (puzzleRecord != null &&
				sizeX <= MAX_CUBES &&
				sizeY <= MAX_CUBES &&
				sizeZ <= MAX_CUBES &&
				(counter < puzzleRecord[l][i][j][k] ||
				puzzleRecord[l][i][j][k] <= NEVER)) {
			readScoreRecords(); /* Maybe its been updated by another */
			puzzleRecord[l][i][j][k] = counter;
			if ((sizeX <= 2 && sizeY <= 2) ||
			    (sizeY <= 2 && sizeZ <= 2) ||
			    (sizeZ <= 2 && sizeX <= 2) ||
			    (orient &&
			    (counter < puzzleRecord[notL][i][j][k] ||
			    puzzleRecord[notL][i][j][k] <= NEVER))) {
				puzzleRecord[notL][i][j][k] = counter;
			}
			writeScore();
			printRecord(sizeX, sizeY, sizeZ,
				orient, practice);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
		}
	}

	class SizeXChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = sizeXValue;

			sizeXValue = sizeXChanger.getValue();
			if (oldValue != sizeXValue) {
				puzzle2D.setRubikSizeX(sizeXValue);
				puzzle3D.setRubikSizeX(sizeXValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeXValue, sizeYValue, sizeZValue,
					orientValue, practiceValue);
			}
		}
	}

	class SizeYChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = sizeYValue;

			sizeYValue = sizeYChanger.getValue();
			if (oldValue != sizeYValue) {
				puzzle2D.setRubikSizeY(sizeYValue);
				puzzle3D.setRubikSizeY(sizeYValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeXValue, sizeYValue, sizeZValue,
					orientValue, practiceValue);
			}
		}
	}

	class SizeZChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = sizeZValue;

			sizeZValue = sizeZChanger.getValue();
			if (oldValue != sizeZValue) {
				puzzle2D.setRubikSizeZ(sizeZValue);
				puzzle3D.setRubikSizeZ(sizeZValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeXValue, sizeYValue, sizeZValue,
					orientValue, practiceValue);
			}
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle2D.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
				puzzle3D.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	class OrientItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = orientValue;

			orientValue = orientCheckBox.isSelected();
			if (oldValue != orientValue) {
				puzzle2D.setRubikOrient(orientValue);
				puzzle3D.setRubikOrient(orientValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeXValue, sizeYValue, sizeZValue,
					orientValue, practiceValue);
			}
		}
	}

	class PracticeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = practiceValue;

			practiceValue = practiceCheckBox.isSelected();
			if (oldValue != practiceValue) {
				puzzle2D.setRubikPractice(practiceValue);
				puzzle3D.setRubikPractice(practiceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				if (!practiceValue)
					messageLabel.setText(
						"Randomize to start");
				printRecord(sizeXValue, sizeYValue, sizeZValue,
					orientValue, practiceValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public void callback(int reason, int dim,
			int face, int position, int direction) {
		messageDisplay = "                                                                                ";
		switch(reason) {
		case RubikInterface.ACTION_MOVED:
			movesDisplay++;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction);
			}
			break;
		case RubikInterface.ACTION_CONTROL:
			if (dim == 2) {
				puzzle3D.movePieces(face, position, direction);
			} else {
				puzzle2D.movePieces(face, position, direction);
			}
			break;
		case RubikInterface.ACTION_UNDO:
			movesDisplay--;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction);
			}
			break;
		case RubikInterface.ACTION_REDO:
			movesDisplay++;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction);
			}
			break;
		case RubikInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				if (dim == 2) {
					delay = puzzle2D.getDelay();
				} else {
					delay = puzzle3D.getDelay();
				}
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					if (dim == 2) {
						puzzle2D.setDelay(delay);
					} else {
						puzzle3D.setDelay(delay);
					}
				}
				if (speedChanger != null)
					speedChanger.setValue(MAX_SPEED +
						MIN_SPEED - 1 - delay);
			}
			return;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public void callback(int reason, int dim) {
		messageDisplay = "";
		switch(reason) {
		case RubikInterface.ACTION_CLEAR:
			if (dim == 2) {
				puzzle3D.resetPieces();
				puzzle3D.drawAllPieces();
			} else {
				puzzle2D.resetPieces();
				puzzle2D.drawAllPieces();
			}
			movesDisplay = 0;
			break;
		case RubikInterface.ACTION_RESTORE:
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			movesDisplay = 0;
			break;
		case RubikInterface.ACTION_RESET:
			movesDisplay = 0;
			break;
		case RubikInterface.ACTION_AMBIGUOUS:
			messageDisplay = "Ambiguous move";
			break;
		case RubikInterface.ACTION_ILLEGAL:
			if (practiceValue ||
			    puzzle2D.checkStarted() || puzzle3D.checkStarted())
				messageDisplay = "Illegal move";
			else
				messageDisplay = "Randomize to start";
			break;
		case RubikInterface.ACTION_MOVED:
			movesDisplay++;
			puzzle2D.setStart(true);
			puzzle3D.setStart(true);
			break;
		case RubikInterface.ACTION_SOLVED:
			if (practiceValue)
				movesDisplay = 0;
			else {
				boolean cheat;

				if (dim == 2) {
					cheat = puzzle2D.getCheat();
				} else {
					cheat = puzzle3D.getCheat();
				}
				if (cheat)
					messageDisplay = "No cheating!!"; //username
				else if (handleSolved(movesDisplay, sizeXValue,
						sizeYValue, sizeZValue,
						orientValue, practiceValue))
					messageDisplay = "Best solution, congratulations!!"; //username
				else
					messageDisplay = "Solved!";
			}
			if (dim == 2) {
				puzzle3D.setStart(false);
			} else {
				puzzle2D.setStart(false);
			}
			//showMessage(messageDisplay);
			break;
		case RubikInterface.ACTION_COMPUTED:
			if (dim == 2) {
				puzzle3D.setStart(false);
			} else {
				puzzle2D.setStart(false);
			}
			break;
		case RubikInterface.ACTION_CHEAT:
			if (dim == 2) {
				puzzle3D.setCheat(puzzle2D.getCheat());
			} else {
				puzzle2D.setCheat(puzzle3D.getCheat());
			}
			break;
		case RubikInterface.ACTION_PRACTICE:
			movesDisplay = 0;
			practiceValue = !practiceValue;
			if (!practiceValue)
				messageDisplay = "Randomize to start";
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikPractice(practiceValue);
			puzzle3D.setRubikPractice(practiceValue);
			practiceCheckBox.setSelected(practiceValue);
			puzzle2D.setStart(false);
			puzzle3D.setStart(false);
			break;
		case RubikInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			practiceValue = false;
			puzzle2D.setRubikPractice(practiceValue);
			puzzle3D.setRubikPractice(practiceValue);
			puzzle2D.setStart(false);
			puzzle3D.setStart(false);
			break;
		case RubikInterface.ACTION_ORIENTIZE:
			movesDisplay = 0;
			orientValue = !orientValue;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikOrient(orientValue);
			puzzle3D.setRubikOrient(orientValue);
			orientCheckBox.setSelected(orientValue);
			break;
		case RubikInterface.ACTION_INCX:
			movesDisplay = 0;
			sizeXValue++;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeX(sizeXValue);
			puzzle3D.setRubikSizeX(sizeXValue);
			if (sizeXChanger != null)
				sizeXChanger.setValue(sizeXValue);
			break;
		case RubikInterface.ACTION_DECX:
			movesDisplay = 0;
			sizeXValue--;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeX(sizeXValue);
			puzzle3D.setRubikSizeX(sizeXValue);
			if (sizeXChanger != null)
				sizeXChanger.setValue(sizeXValue);
			break;
		case RubikInterface.ACTION_INCY:
			movesDisplay = 0;
			sizeYValue++;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeY(sizeYValue);
			puzzle3D.setRubikSizeY(sizeYValue);
			if (sizeYChanger != null)
				sizeYChanger.setValue(sizeYValue);
			break;
		case RubikInterface.ACTION_DECY:
			movesDisplay = 0;
			sizeYValue--;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeY(sizeYValue);
			puzzle3D.setRubikSizeY(sizeYValue);
			if (sizeYChanger != null)
				sizeYChanger.setValue(sizeYValue);
			break;
		case RubikInterface.ACTION_INCZ:
			movesDisplay = 0;
			sizeZValue++;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeZ(sizeZValue);
			puzzle3D.setRubikSizeZ(sizeZValue);
			if (sizeZChanger != null)
				sizeZChanger.setValue(sizeZValue);
			break;
		case RubikInterface.ACTION_DECZ:
			movesDisplay = 0;
			sizeZValue--;
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			puzzle2D.setRubikSizeZ(sizeZValue);
			puzzle3D.setRubikSizeZ(sizeZValue);
			if (sizeZChanger != null)
				sizeZChanger.setValue(sizeZValue);
			break;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public RubikFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel checkBoxPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel checkBoxGroupPanel = new JPanel();
		JPanel canvasPanel = new JPanel();
		JPanel canvas2Panel = new JPanel();
		JPanel canvas3Panel = new JPanel();

		try {
			mediumIcon = getIcon(RubikInterface.ICONS[RubikInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(RubikInterface.ICONS[RubikInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			RubikInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Practice");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Increment");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_I);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Decrement");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_D);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Orientize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sound @");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Randomize"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(8);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			sliderPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			sizeXChanger = new Spinner("SizeX",
				sizeXValue, RubikInterface.MIN_CUBES);
			sizeYChanger = new Spinner("SizeY",
				sizeYValue, RubikInterface.MIN_CUBES);
			sizeZChanger = new Spinner("SizeZ",
				sizeZValue, RubikInterface.MIN_CUBES);
			/*sizeXChanger = new Slider("SizeX",
				sizeXValue, RubikInterface.MIN_CUBES, MAX_CUBES);
			sizeYChanger = new Slider("SizeY",
				sizeYValue, RubikInterface.MIN_CUBES, MAX_CUBES);
			sizeZChanger = new Slider("SizeZ",
				sizeZValue, RubikInterface.MIN_CUBES, MAX_CUBES);*/
			sliderPanel.add(sizeXChanger);
			sliderPanel.add(sizeYChanger);
			sliderPanel.add(sizeZChanger);
			checkBoxGroupPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			checkBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			checkBoxPanel.add(orientCheckBox);
			orientCheckBox.setSelected(orientValue);
			checkBoxPanel.add(practiceCheckBox);
			practiceCheckBox.setSelected(practiceValue);
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			speedChanger.setWidth(150);*/
			checkBoxPanel.add(speedChanger);
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			topPanel.add(sliderPanel);
			topPanel.add(checkBoxPanel);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		puzzle2D = new Rubik(this, 2,
			sizeXValue, sizeYValue, sizeZValue,
			orientValue, practiceValue,
			false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3], faceColor[4], faceColor[5],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		puzzle3D = new Rubik(this, 3,
			sizeXValue, sizeYValue, sizeZValue,
			orientValue, practiceValue,
			false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3], faceColor[4], faceColor[5],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		canvas2Panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvas3Panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvas2Panel.add(puzzle2D);
		canvas3Panel.add(puzzle3D);
		JPanel extraPanel = new JPanel();
		canvasPanel.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.gridx = 0;
		c.gridy = 0;
		canvasPanel.add(canvas2Panel, c);
		c.gridx = 0;
		c.gridy = 1;
		canvasPanel.add(canvas3Panel, c);
		extraPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		extraPanel.add(canvasPanel);
		secondaryPanel.add(extraPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(sizeXValue, sizeYValue, sizeZValue,
				orientValue, practiceValue);
			sizeXChanger.addChangeListener(new SizeXChangeListener());
			sizeYChanger.addChangeListener(new SizeYChangeListener());
			sizeZChanger.addChangeListener(new SizeZChangeListener());
			orientCheckBox.addItemListener(new OrientItemListener());
			practiceCheckBox.addItemListener(new PracticeItemListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
		}
		canvas2Panel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle2D.requestFocus();
			}
		});
		canvas3Panel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle3D.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle2D.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle2D.setSize(widthValue, heightValue);
		/* puzzle3D.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle3D.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : RubikInterface.ICONS) {
			URL url = RubikFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-faceColor{0|1|2|3|4|5}={int}]" + NL +
"\t[-borderColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}] [-size{x|y|z}={int}]" + NL +
"\t[-orient={bool}] [-practice={bool}] [-delay={int}]");
			return;
		}
		RubikFrame frame = new RubikFrame(RubikInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle3D.requestFocusInWindow();
		frame.setLocation(RubikFrame.xValue, RubikFrame.yValue);
		frame.setVisible(true);
	}
}
