package com.sillycycle.bagleyd.rubik;

/*
 * @(#)RubikSolve.java
 *
 * Taken from code originally written by
 * Michael B. Martin <martinm@sps1.phys.vt.edu>
 * From cubist10.c-- for IBM PC.
 * Used by permission.
 * Taken from the algorithm in the Ideal Solution book.
 * Revenge stuff taken from Rubik's Revenge Puzzle The Ideal Solution
 * book 1982.
 *
 * Copyright 1999 - 2021  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import javax.swing.JFrame;

/**
 * The <code>RubikSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class RubikSolve {
	private static JFrame frame;
	private static Rubik rubik;
	//private static int counter; // debugging

	RubikSolve(JFrame frame, Rubik rubik) {
		RubikSolve.frame = frame;
		RubikSolve.rubik = rubik;
		//this.counter = 0;
	}

	static final int TOP_FACE = 0; // BLUE
	static final int LEFT_FACE = 1; // WHITE
	static final int FRONT_FACE = 2; // RED
	static final int RIGHT_FACE = 3; // YELLOW
	static final int BOTTOM_FACE = 4; // GREEN
	static final int BACK_FACE = 5; // ORANGE

	boolean moveAPiece(int face, int position, int direction,
			boolean control) {
		if (!getFlag()) {
			return false;
		}
		rubik.movePuzzleDelay(face, position, direction,
			control, Rubik.DOUBLE);
		return true;
	}

	/* Mappings of Ideal notation to my General nxnxn cube notation */
	void rotateLeft() {
		moveAPiece(FRONT_FACE, 0, Rubik.LEFT, true);
	}

	void rotateRight() {
		moveAPiece(FRONT_FACE, 0, Rubik.RIGHT, true);
	}

	void rotateUp() {
		moveAPiece(FRONT_FACE, 0, Rubik.TOP, true);
	}

	void rotateDown() {
		moveAPiece(FRONT_FACE, 0, Rubik.BOTTOM, true);
	}

	void rotateCw() {
		moveAPiece(TOP_FACE, 0, Rubik.RIGHT, true);
	}

	void rotateCcw() {
		moveAPiece(TOP_FACE, 0, Rubik.LEFT, true);
	}

	void rotateTopLeft() {
		moveAPiece(FRONT_FACE, 0, Rubik.LEFT, false);
	}

	void rotateCenterLeft() {
		moveAPiece(FRONT_FACE, rubik.sizeX * (rubik.sizeX >> 1),
			Rubik.LEFT, false);
	}

	void rotateBottomLeft() {
		moveAPiece(FRONT_FACE, rubik.sizeX * (rubik.sizeX - 1),
			Rubik.LEFT, false);
	}

	void rotateTopRight() {
		moveAPiece(FRONT_FACE, 0, Rubik.RIGHT, false);
	}

	void rotateCenterRight() {
		moveAPiece(FRONT_FACE, rubik.sizeX * (rubik.sizeX >> 1),
			Rubik.RIGHT, false);
	}

	void rotateBottomRight() {
		moveAPiece(FRONT_FACE, rubik.sizeX * (rubik.sizeX - 1),
			Rubik.RIGHT, false);
	}

	void rotateLeftUp() {
		moveAPiece(FRONT_FACE, 0, Rubik.TOP, false);
	}

	void rotateCenterUp() {
		moveAPiece(FRONT_FACE, rubik.sizeX >> 1, Rubik.TOP, false);
	}

	void rotateRightUp() {
		moveAPiece(FRONT_FACE, rubik.sizeX - 1, Rubik.TOP, false);
	}

	void rotateLeftDown() {
		moveAPiece(FRONT_FACE, 0, Rubik.BOTTOM, false);
	}

	void rotateCenterDown() {
		moveAPiece(FRONT_FACE, rubik.sizeX >> 1, Rubik.BOTTOM, false);
	}

	void rotateRightDown() {
		moveAPiece(FRONT_FACE, rubik.sizeX - 1, Rubik.BOTTOM, false);
	}

	void rotateFrontCw() {
		moveAPiece(TOP_FACE, rubik.sizeX * (rubik.sizeX - 1),
			Rubik.RIGHT, false);
	}

	void rotateFrontCcw() {
		moveAPiece(TOP_FACE, rubik.sizeX * (rubik.sizeX - 1),
			Rubik.LEFT, false);
	}

	void rotateBackCw() {
		moveAPiece(TOP_FACE, 0, Rubik.LEFT, false);
	}

	void rotateBackCcw() {
		moveAPiece(TOP_FACE, 0, Rubik.RIGHT, false);
	}

	static final int[/*MAX_FACES*/][/*MAX_ORIENT*/] otherFaceFromEdge =
	{
		{5, 3, 2, 1},
		{0, 2, 4, 5},
		{0, 3, 4, 1},
		{0, 5, 4, 2},
		{2, 3, 5, 1},
		{4, 3, 0, 1}
	};

	static final int[/*MAX_FACES*/][/*2 * MAX_ORIENT*/] otherEdgeFromEdge =
	{
		{5, 4, 1, 0, 1, 0, 1, 0},
		{7, 6, 7, 6, 7, 6, 7, 6},
		{5, 4, 7, 6, 1, 0, 3, 2},
		{3, 2, 3, 2, 3, 2, 3, 2},
		{5, 4, 5, 4, 1, 0, 5, 4},
		{5, 4, 3, 2, 1, 0, 7, 6}
	};

	/* clock is a little distorted */
	static final int[/*MAX_ORIENT * 2*/] clockEdgeToPosition =
	{1, 2, 7, 11, 14, 13, 8, 4};

	static final int[/*16*/] positionToClockEdge =
	{  -1, 0, 1, -1,
	   7, -1, -1, 2,
	   6, -1, -1, 3,
	   -1, 5, 4, -1
	};

	static final int[/*MAX_ORIENT*/] clockCenterToPosition =
	{5, 6, 10, 9};

	static final int[/*MAX_FACES*/] oppositeFace =
	{BOTTOM_FACE, RIGHT_FACE, BACK_FACE, LEFT_FACE, TOP_FACE, FRONT_FACE};

	static final int[/*MAX_ORIENT*/] bestOrder =
	{0, 1, 3, 2};

	static final int[/*MAX_ORIENT*/] lastColumn =
	{
		BACK_FACE, BOTTOM_FACE, FRONT_FACE, TOP_FACE
	};

	static final int[/*MAX_FACES*/] offsetCenters =
	{3, 0, 2, 3, 0, 0};

	/*static final int[MAX_FACES] mapGeneralToIdeal =
		{5, 4, 1, 2, 6, 3}; */

	static final int[/*MAX_FACES*/] mapIdealToGeneral =
		{2, 3, 5, 1, 0, 4}; /* Remember to subtract 1 */

	static int cornerFace(int quad) {
		/* top left to bottom right */
		switch (quad) {
		case 0:
			return 0;
		case 1:
			return rubik.sizeX - 1;
		case 2:
			return rubik.sizeX * rubik.sizeX
				- rubik.sizeX;
		case 3:
			return rubik.sizeX * rubik.sizeX - 1;
		default:
			return rubik.sizeX * rubik.sizeX;
		}
	}

	static int sideFace(int quad) {
		/* top left to bottom right */
		/* if even return lowest */
		switch (quad) {
		case 0:
			return ((rubik.sizeX - 1) >> 1);
		case 1:
			return ((rubik.sizeX - 1) >> 1) * rubik.sizeX;
		case 2:
			return ((rubik.sizeX - 1) >> 1) * rubik.sizeX
				+ rubik.sizeX - 1;
		case 3:
			return rubik.sizeX * rubik.sizeX
				- (rubik.sizeX >> 1) - 1;
		default:
			return rubik.sizeX * rubik.sizeX;
		}
	}

	static int centerFace() {
		/* if even return lowest */
		return ((rubik.sizeX - 1) >> 1) * rubik.sizeX
			+ ((rubik.sizeX - 1) >> 1);
	}

	static int colorSide(int m, int c) {
		int d, i, j;

		d = mapIdealToGeneral[m - 1];
		i = c % rubik.sizeX;
		j = c / rubik.sizeX;
		return rubik.cubeLoc[d][j * rubik.sizeX + i].face;
	}

	static int mapFaceOrient(int m, int pos) {
		int d;

		d = mapIdealToGeneral[m - 1];
		return ((rubik.cubeLoc[d][pos].rotation -
			rubik.cubeLoc[d][0].rotation + RubikInterface.MAX_ORIENT)
				% RubikInterface.MAX_ORIENT);
	}

	static int mapFaceOrientRev(int m, int pos) {
		return ((rubik.cubeLoc[m][pos].rotation -
			rubik.cubeLoc[m][0].rotation + RubikInterface.MAX_ORIENT)
				% RubikInterface.MAX_ORIENT);
	}

	static int mapFace(int m) {
		return mapFaceOrient(m, centerFace());
	}

	/* This procedure finds the location of a specified corner.  An */
	/* improperly-specified color combination will result in a return */
	/* value of 9. */
	int findCorner(int color1, int color2, int color3) {
		int corner = 0, temp = 1;

		do {
			if (colorSide(5, cornerFace(2)) == color1) {
				if (colorSide(1, cornerFace(0)) == color2 &&
						colorSide(4, cornerFace(1)) == color3)
					corner = temp;
			} else if (colorSide(5, cornerFace(2)) == color2) {
				if (colorSide(1, cornerFace(0)) == color3 &&
						colorSide(4, cornerFace(1)) == color1)
					corner = temp;
			} else if (colorSide(5, cornerFace(2)) == color3 &&
					colorSide(1, cornerFace(0)) == color1 &&
					colorSide(4, cornerFace(1)) == color2) {
				corner = temp;
			}
			if (corner == 0) {
				if (temp < 4)
					rotateLeft();
				else if (temp == 4) {
					rotateCw();
					rotateCw();
				} else if (temp < 8)
					rotateRight();
				else if (temp == 8)
					corner = 9;
				temp++;
			}
		} while (corner == 0);

		/* put the cube back to the way it was */
		switch (corner) {
		case 2:
			rotateRight();
			break;
		case 3:
			rotateRight();
			rotateRight();
			break;
		case 4:
			rotateLeft();
			break;
		case 5:
			rotateCw();
			rotateCw();
			rotateLeft();
			break;
		case 6:
			rotateCw();
			rotateCw();
			break;
		case 7:
			rotateCw();
			rotateCw();
			rotateRight();
			break;
		case 8:
			rotateCw();
			rotateCw();
			rotateRight();
			rotateRight();
			break;
		default:
			break;
		}
		return (corner);
	}

	int offsetRevenge(int position) {
		int x = position % 4;
		int y = position / 4;

		if (x > 1)
			x += rubik.sizeX - 4;
		if (y > 1)
			y += rubik.sizeX - 4;
		return x + y * rubik.sizeX;
	}

	int revengeOffset(int position) {
		int x = position % rubik.sizeX;
		int y = position / rubik.sizeX;

		if (x > (rubik.sizeX >> 1))
			x += 4 - rubik.sizeX;
		if (y > 1)
			y += 4 - rubik.sizeX;
		return x + y * 4;
	}

	RubikFE findClockEdge(int matchFace, int matchEdge) {
		int face, edge;
		RubikFE rubikFE = new RubikFE();

		for (face = 0; face < Rubik.MAX_FACES; face++) {
			for (edge = 0; edge < 2 * RubikInterface.MAX_ORIENT; edge++) {
				int position, otherPosition, otherFace, otherEdge;

				position = clockEdgeToPosition[edge];
				position = offsetRevenge(position); /* non revenge */
				otherFace = otherFaceFromEdge[face][edge >> 1];
				otherEdge = otherEdgeFromEdge[face][edge];
				otherPosition = clockEdgeToPosition[otherEdge];
				otherPosition = offsetRevenge(otherPosition); /* non revenge */
				if (matchFace == rubik.cubeLoc[face][position].face &&
						(matchEdge & 1) == (edge & 1) &&
						otherFaceFromEdge[matchFace][matchEdge >> 1] ==
						rubik.cubeLoc[otherFace][otherPosition].face) {
					position = revengeOffset(position); /* back to revenge */
					rubikFE.face = face;
					rubikFE.edge = positionToClockEdge[position];
					/*rubikFP.x = position % rubik.sizeX;
					rubikFP.y = position / rubik.sizeX;*/
					return rubikFE;
				}
			}
		}
		rubikFE.face = -1;
		rubikFE.edge = -1;
		/*rubikFP.x = -1;
		rubikFP.y = -1;*/
		return rubikFE;
	}

	boolean matchEdges(int matchFace0, int matchFace1, int matchEdge) {
		int clockEdge0 = matchEdge << 1;
		int clockEdge1 = clockEdge0 + 1;
		int position0, position1, otherPosition0, otherPosition1;
		int otherFace0, otherFace1, otherClockEdge0, otherClockEdge1;

		position0 = clockEdgeToPosition[clockEdge0];
		position0 = offsetRevenge(position0); /* non revenge */
		position1 = clockEdgeToPosition[clockEdge1];
		position1 = offsetRevenge(position1); /* non revenge */
		otherFace0 = otherFaceFromEdge[matchFace0][matchEdge];
		otherFace1 = otherFaceFromEdge[matchFace1][matchEdge];
		otherClockEdge0 = otherEdgeFromEdge[matchFace0][clockEdge0];
		otherClockEdge1 = otherEdgeFromEdge[matchFace1][clockEdge1];
		otherPosition0 = clockEdgeToPosition[otherClockEdge0];
		otherPosition0 = offsetRevenge(otherPosition0); /* non revenge */
		otherPosition1 = clockEdgeToPosition[otherClockEdge1];
		otherPosition1 = offsetRevenge(otherPosition1); /* non revenge */
		return (rubik.cubeLoc[matchFace0][position0].face ==
			rubik.cubeLoc[matchFace1][position1].face &&
			rubik.cubeLoc[otherFace0][otherPosition0].face ==
			rubik.cubeLoc[otherFace1][otherPosition1].face);
	}


	/* This procedure finds the location of a specified edge.  An */
	/* improperly-specified color combination will result in a return */
	/* value of 13. */
	int findEdge(int color1, int color2) {
		int edge = 0, temp = 1;

		do {
			/* Top (5) down & Front (1) top edge */
			if (colorSide(5, sideFace(3)) == color1) {
				if (colorSide(1, sideFace(0)) == color2)
					edge = temp;
			} else if (colorSide(5, sideFace(3)) == color2 &&
					colorSide(1, sideFace(0)) == color1) {
				edge = temp;
			}
			if (edge == 0) {
				if (temp < 4)
					rotateLeft();
				else if (temp == 4) {
					rotateLeft();
					rotateCw();
				} else if (temp < 8)
					rotateUp();
				else if (temp == 8) {
					rotateUp();
					rotateCw();
				} else if (temp < 12)
					rotateRight();
				else if (temp == 12)
					edge = 13;	/* end condition, just in case */
				temp++;
			}
		} while (edge == 0);

		/* put the cube back to the way it was */
		switch (edge) {
		case 2:
			rotateRight();
			break;
		case 3:
			rotateRight();
			rotateRight();
			break;
		case 4:
			rotateLeft();
			break;
		case 5:
			rotateCcw();
			break;
		case 6:
			rotateCcw();
			rotateRight();
			break;
		case 7:
			rotateCcw();
			rotateRight();
			rotateRight();
			break;
		case 8:
			rotateCcw();
			rotateLeft();
			break;
		case 9:
			rotateCcw();
			rotateCcw();
			break;
		case 10:
			rotateCcw();
			rotateCcw();
			rotateRight();
			break;
		case 11:
			rotateUp();
			rotateUp();
			break;
		case 12:
			rotateUp();
			rotateUp();
			rotateRight();
			break;
		default:
			break;
		}
		return (edge);
	}

	/* This procedure places the specified edge piece in edge position */
	/* #1 (front top). */
	void placeEdge(int edge, int topColor) {
		/* first put the edge piece in position */
		/* #8 (center row, left rear) */
		switch (edge) {
		case 1:
			if (colorSide(5, sideFace(3)) == topColor)
				return;	/* already in place */
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCw();
			break;
		case 2:
			rotateTopLeft();
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCw();
			rotateTopRight();
			break;
		case 3:
			rotateTopLeft();
			rotateTopLeft();
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCw();
			rotateTopRight();
			rotateTopRight();
			break;
		case 4:
			rotateTopRight();
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCw();
			rotateTopLeft();
			break;
		case 5:
			rotateCenterLeft();
			break;
		case 6:
			rotateCenterLeft();
			rotateCenterLeft();
			break;
		case 7:
			rotateCenterRight();
			break;
		case 9:
			rotateFrontCw();
			rotateCenterLeft();
			rotateFrontCcw();
			break;
		case 10:
			rotateBottomLeft();
			rotateFrontCw();
			rotateCenterLeft();
			rotateFrontCcw();
			rotateBottomRight();
			break;
		case 11:
			rotateBottomLeft();
			rotateBottomLeft();
			rotateFrontCw();
			rotateCenterLeft();
			rotateFrontCcw();
			rotateBottomRight();
			rotateBottomRight();
			break;
		case 12:
			rotateBottomRight();
			rotateFrontCw();
			rotateCenterLeft();
			rotateFrontCcw();
			rotateBottomLeft();
			break;
		default:
			break;
		}
		/* put the piece in place */
		if (colorSide(4, sideFace(1)) == topColor) {
			rotateFrontCw();
			rotateCenterRight();
			rotateCenterRight();
			rotateFrontCcw();
		} else {
			rotateFrontCcw();
			rotateCenterRight();
			rotateFrontCw();
		}
	}

	/* This procedure solves the top (TOP_FACE) side, except for one edge. */
	void solveTop() {
		int corner, edge;

		if ((rubik.sizeX & 1) == 1) {
			/* put the blue face on the top */
			if (colorSide(1, centerFace()) == TOP_FACE)
				rotateUp();
			else if (colorSide(2, centerFace()) == TOP_FACE)
				rotateCcw();
			else if (colorSide(3, centerFace()) == TOP_FACE)
				rotateDown();
			else if (colorSide(4, centerFace()) == TOP_FACE)
				rotateCw();
			else if (colorSide(6, centerFace()) == TOP_FACE) {
				rotateUp();
				rotateUp();
			}
		}
		/* first find the blue-red-white corner and place it */
		corner = findCorner(TOP_FACE, FRONT_FACE, LEFT_FACE);
		switch (corner) {
		case 1:
			if (colorSide(5, cornerFace(2)) == FRONT_FACE) {
				rotateFrontCw();
				rotateTopLeft();
			} else if (colorSide(5, cornerFace(2)) == LEFT_FACE) {
				rotateLeftUp();
				rotateTopRight();
			}
			break;
		case 2:
			switch (colorSide(5, cornerFace(3))) {
			case TOP_FACE:
				rotateTopLeft();
				break;
			case FRONT_FACE:
				rotateRightUp();
				rotateTopLeft();
				rotateTopLeft();
				break;
			case LEFT_FACE:
				rotateFrontCcw();
				break;
			default:
				break;
			}
			break;
		case 3:
			switch (colorSide(5, cornerFace(1))) {
			case TOP_FACE:
				rotateTopLeft();
				rotateTopLeft();
				break;
			case FRONT_FACE:
				rotateTopRight();
				rotateLeftDown();
				break;
			case LEFT_FACE:
				rotateRightDown();
				rotateTopLeft();
				break;
			default:
				break;
			}
			break;
		case 4:
			switch (colorSide(5, cornerFace(0))) {
			case TOP_FACE:
				rotateTopRight();
				break;
			case FRONT_FACE:
				rotateLeftDown();
				break;
			case LEFT_FACE:
				rotateTopRight();
				rotateLeftUp();
				rotateTopRight();
				break;
			default:
				break;
			}
			break;
		case 5:
			switch (colorSide(6, cornerFace(0))) {
			case TOP_FACE:
				rotateBottomLeft();
				rotateLeftUp();
				rotateLeftUp();
				break;
			case FRONT_FACE:
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 6:
			switch (colorSide(6, cornerFace(1))) {
			case TOP_FACE:
				rotateFrontCw();
				rotateFrontCw();
				break;
			case FRONT_FACE:
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 7:
			switch (colorSide(6, cornerFace(3))) {
			case TOP_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case FRONT_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 8:
			switch (colorSide(6, cornerFace(2))) {
			case TOP_FACE:
				rotateLeftUp();
				rotateLeftUp();
				break;
			case FRONT_FACE:
				rotateBottomRight();
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		/* now find the blue-yellow-red corner and place it */
		rotateLeft();
		corner = findCorner(TOP_FACE, RIGHT_FACE, FRONT_FACE);
		switch (corner) {
		case 1:
			if (colorSide(5, cornerFace(2)) == RIGHT_FACE) {
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCcw();
				rotateFrontCcw();
			} else if (colorSide(5, cornerFace(2)) == FRONT_FACE) {
				rotateFrontCw();
				rotateFrontCw();
				rotateBottomLeft();
				rotateFrontCw();
			}
			break;
		case 2:
			switch (colorSide(5, cornerFace(3))) {
			case TOP_FACE:
				rotateRightDown();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateRightDown();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case FRONT_FACE:
				rotateFrontCcw();
				break;
			default:
				break;
			}
			break;
		case 3:
			switch (colorSide(5, cornerFace(1))) {
			case TOP_FACE:
				rotateRightDown();
				rotateRightDown();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateRightDown();
				rotateFrontCcw();
				break;
			case FRONT_FACE:
				rotateRightDown();
				rotateRightDown();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 5:
			switch (colorSide(6, cornerFace(0))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case FRONT_FACE:
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 6:
			switch (colorSide(6, cornerFace(1))) {
			case TOP_FACE:
				rotateFrontCw();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateRightUp();
				rotateFrontCcw();
				break;
			case FRONT_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 7:
			switch (colorSide(6, cornerFace(3))) {
			case TOP_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateBottomLeft();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case FRONT_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 8:
			switch (colorSide(6, cornerFace(2))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateBottomRight();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateBottomRight();
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case FRONT_FACE:
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		/* now find the blue-orange-yellow corner and place it */
		rotateLeft();
		corner = findCorner(TOP_FACE, BACK_FACE, RIGHT_FACE);
		switch (corner) {
		case 1:
			if (colorSide(5, cornerFace(2)) == BACK_FACE) {
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCcw();
				rotateFrontCcw();
			} else if (colorSide(5, cornerFace(2)) == RIGHT_FACE) {
				rotateFrontCw();
				rotateFrontCw();
				rotateBottomLeft();
				rotateFrontCw();
			}
			break;
		case 2:
			switch (colorSide(5, cornerFace(3))) {
			case TOP_FACE:
				rotateRightDown();
				rotateBottomLeft();
				rotateRightUp();
				rotateFrontCw();
				break;
			case BACK_FACE:
				rotateFrontCw();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			case RIGHT_FACE:
				rotateFrontCcw();
				break;
			default:
				break;
			}
			break;
		case 5:
			switch (colorSide(6, cornerFace(0))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case BACK_FACE:
				rotateRightDown();
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case RIGHT_FACE:
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 6:
			switch (colorSide(6, cornerFace(1))) {
			case TOP_FACE:
				rotateFrontCw();
				rotateFrontCw();
				break;
			case BACK_FACE:
				rotateBottomLeft();
				rotateRightDown();
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case RIGHT_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 7:
			switch (colorSide(6, cornerFace(3))) {
			case TOP_FACE:
				rotateBottomLeft();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case BACK_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateRightDown();
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case RIGHT_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 8:
			switch (colorSide(6, cornerFace(2))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateBottomRight();
				rotateFrontCw();
				rotateFrontCw();
				break;
			case BACK_FACE:
				rotateRightDown();
				rotateBottomRight();
				rotateBottomRight();
				rotateRightUp();
				rotateFrontCcw();
				break;
			case RIGHT_FACE:
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		/* and now find the blue-white-orange corner and place it */
		rotateLeft();
		corner = findCorner(TOP_FACE, LEFT_FACE, BACK_FACE);
		switch (corner) {
		case 1:
			if (colorSide(5, cornerFace(2)) == LEFT_FACE) {
				rotateLeftDown();
				rotateBottomRight();
				rotateLeftUp();
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
			} else if (colorSide(5, cornerFace(2)) == BACK_FACE) {
				rotateFrontCcw();
				rotateBottomLeft();
				rotateFrontCw();
				rotateBottomRight();
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
			}
			break;
		case 5:
			switch (colorSide(6, cornerFace(0))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateBottomLeft();
				rotateLeftUp();
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case BACK_FACE:
				rotateBottomLeft();
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 6:
			switch (colorSide(6, cornerFace(1))) {
			case TOP_FACE:
				rotateBottomRight();
				rotateFrontCcw();
				rotateBottomLeft();
				rotateFrontCw();
				rotateBottomLeft();
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
				break;
			case LEFT_FACE:
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case BACK_FACE:
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 7:
			switch (colorSide(6, cornerFace(3))) {
			case TOP_FACE:
				rotateLeftDown();
				rotateBottomRight();
				rotateLeftUp();
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case LEFT_FACE:
				rotateLeftDown();
				rotateBottomLeft();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case BACK_FACE:
				rotateFrontCcw();
				rotateBottomLeft();
				rotateBottomLeft();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		case 8:
			switch (colorSide(6, cornerFace(2))) {
			case TOP_FACE:
				rotateFrontCcw();
				rotateBottomRight();
				rotateBottomRight();
				rotateFrontCw();
				rotateBottomLeft();
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
				break;
			case LEFT_FACE:
				rotateBottomRight();
				rotateBottomRight();
				rotateLeftDown();
				rotateBottomLeft();
				rotateLeftUp();
				break;
			case BACK_FACE:
				rotateFrontCcw();
				rotateBottomRight();
				rotateFrontCw();
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		if ((rubik.sizeX & 1) == 1) {
			rotateLeft();

			/* find the blue-red edge and place it */
			edge = findEdge(TOP_FACE, FRONT_FACE);
			placeEdge(edge, TOP_FACE);
			rotateLeft();

			/* find the blue-yellow edge and place it */
			edge = findEdge(TOP_FACE, RIGHT_FACE);
			placeEdge(edge, TOP_FACE);
			rotateLeft();

			/* find the blue-orange edge and place it */
			edge = findEdge(TOP_FACE, BACK_FACE);
			placeEdge(edge, TOP_FACE);
			rotateLeft();
		}
		rotateUp(); /* put the blue side to the back */
	}

	/* This procedure swaps the the bottom front corners. */
	void swapBottomCorners() {
		rotateTopRight();
		rotateFrontCw();
		rotateTopLeft();
		rotateLeftUp();
		rotateTopLeft();
		rotateLeftDown();
		rotateTopRight();
		rotateFrontCw();
		rotateFrontCw();
	}

	/* This procedure places the front (BOTTOM_FACE) four corners into */
	/* their correct positions. */
	void alignCorners() {
		int corner;

		/* find and place the green-orange-white corner (upper left) */
		corner = findCorner(BOTTOM_FACE, BACK_FACE, LEFT_FACE);
		switch (corner) {
		case 2:
			rotateFrontCcw();
			break;
		case 5:
			rotateFrontCw();
			break;
		case 6:
			rotateFrontCw();
			rotateFrontCw();
			break;
		default:
			break;
		}
		/* find and place the green-yellow-orange corner */
		/* (lower left) */
		corner = findCorner(BOTTOM_FACE, RIGHT_FACE, BACK_FACE);
		if (corner == 2) {
			rotateCw();
			swapBottomCorners();
			rotateCcw();
			swapBottomCorners();
		} else if (corner == 6)
			swapBottomCorners();

		/* find and place the green-red-yellow corner (lower right) */
		corner = findCorner(BOTTOM_FACE, FRONT_FACE, RIGHT_FACE);
		if (corner == 2) {
			rotateCw();
			swapBottomCorners();
			rotateCcw();
		}
	}

	/* This procedure aligns the bottom corner colors (may need to be */
	/* repeated). */
	void colorAlignFront() {
		rotateTopRight();
		rotateFrontCw();
		rotateFrontCw();
		rotateTopLeft();
		rotateFrontCw();
		rotateTopRight();
		rotateFrontCw();
		rotateTopLeft();
		rotateFrontCw();
		rotateFrontCw();
	}

	/* This procedure moves the remaining TOP_FACE edge piece into position */
	/* from edge position #6. */
	void alignLastEdge() {
		/* edge piece should be in edge position #6 */
		/* check its orientation and decide which sequence to */
		/* perform */
		if (colorSide(1, sideFace(2)) == TOP_FACE) {
			rotateCenterLeft();
			rotateFrontCw();
			rotateCenterRight();
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCcw();
			rotateCenterRight();
			rotateFrontCw();
		} else {
			rotateFrontCcw();
			rotateCenterLeft();
			rotateFrontCw();
			rotateCenterRight();
			rotateFrontCw();
			rotateCenterLeft();
			rotateFrontCcw();
		}
	}

	/* This procedure "shuffles" the three center edges on the front and */
	/* top faces. */
	void shuffleEdges()
	{
		rotateCenterUp();
		rotateTopRight();
		rotateTopRight();
		rotateCenterDown();
		rotateTopRight();
		rotateTopRight();
	}

	/* This procedure should be used when the two opposing top front and */
	/* back edge pieces are in position but not color aligned (this */
	/* sequence is known as Rubik's Maneuver). */
	void recolorTopEdges() {
		rotateCenterUp();
		rotateTopRight();
		rotateCenterUp();
		rotateTopRight();
		rotateCenterUp();
		rotateTopRight();
		rotateTopRight();
		rotateCenterDown();
		rotateTopRight();
		rotateCenterDown();
		rotateTopRight();
		rotateCenterDown();
		rotateTopRight();
		rotateTopRight();
	}

	/* This procedure completes the BOTTOM_FACE side by color-aligning the */
	/* BOTTOM_FACE corners and putting the BOTTOM_FACE edges in place. */
	void solveBottom() {
		int aligned;

		/* the BOTTOM_FACE corners (currently in the front) should now */
		/* be in their proper locations; next, we color align them, and */
		/* then move the bottom edges into place */
		do {
			aligned = 0;
			if (colorSide(1, cornerFace(0)) == BOTTOM_FACE)
				aligned++;
			if (colorSide(1, cornerFace(1)) == BOTTOM_FACE)
				aligned++;
			if (colorSide(1, cornerFace(2)) == BOTTOM_FACE)
				aligned++;
			if (colorSide(1, cornerFace(3)) == BOTTOM_FACE)
				aligned++;

			switch (aligned) {
			case 0:
				colorAlignFront();
				break;
			case 1:
				/* place aligned corner in upper right */
				if (colorSide(1, cornerFace(0)) == BOTTOM_FACE)
					rotateFrontCw();
				if (colorSide(1, cornerFace(2)) == BOTTOM_FACE) {
					rotateFrontCw();
					rotateFrontCw();
				}
				if (colorSide(1, cornerFace(3)) == BOTTOM_FACE)
					rotateFrontCcw();
				colorAlignFront();
				break;
			case 2:
				if ((colorSide(1, cornerFace(0)) != BOTTOM_FACE)
						|| (colorSide(1, cornerFace(1)) == BOTTOM_FACE))
					rotateFrontCw();
				if (colorSide(1, cornerFace(0)) != BOTTOM_FACE)
					rotateFrontCw();
				colorAlignFront();
				break;
			case 3:
				/* not sure if this is possible */
				colorAlignFront();
				break;
			default:
				break;
			}
		} while (aligned < 4);
	}

	void solve_2_2() {
		if (colorSide(5, cornerFace(0)) == colorSide(4, cornerFace(1))) {
			rotateFrontCw();
			return;
		}
		for (int i = 0; i < 2; i++) {
			if (colorSide(5, cornerFace(0)) == colorSide(5, cornerFace(2)))
				return;
			rotateFrontCcw();
		}
	}

	void solveBottomEdges() {
		int edge, color;

		/* next we move the bottom edges into place */
		rotateDown();		/* put the green face on top */
		rotateCw();
		rotateCw();

		color = colorSide(1, cornerFace(0));	/* get upper front corner color */
		edge = findEdge(BOTTOM_FACE, color);
		placeEdge(edge, BOTTOM_FACE);
		rotateTopRight();

		color = colorSide(1, cornerFace(0));	/* get upper front corner color */
		edge = findEdge(BOTTOM_FACE, color);
		placeEdge(edge, BOTTOM_FACE);
		rotateTopRight();

		color = colorSide(1, cornerFace(0));	/* get upper front corner color */
		edge = findEdge(BOTTOM_FACE, color);
		placeEdge(edge, BOTTOM_FACE);
		rotateTopRight();

		color = colorSide(1, cornerFace(0));	/* get upper front corner color */
		edge = findEdge(BOTTOM_FACE, color);
		placeEdge(edge, BOTTOM_FACE);
		rotateTopRight();

		/* now, align the fourth blue edge piece (if necessary) */
		rotateCw();
		rotateCw();
		edge = findEdge(TOP_FACE, LEFT_FACE);
		if (edge == 1) {
			if (colorSide(1, sideFace(0)) == TOP_FACE) {
				rotateFrontCcw();
				rotateCenterLeft();
				rotateFrontCw();
				rotateCenterLeft();
				rotateFrontCw();
				rotateCenterLeft();
				rotateFrontCcw();
			}
		} else {
			switch (edge) {
			case 5:
				rotateCenterRight();
				break;
			case 7:
				rotateCenterLeft();
				break;
			case 8:
				rotateCenterRight();
				rotateCenterRight();
				break;
			default:
				break;
			}
			alignLastEdge();
		}
	}

	/* This procedure completes the solution process by placing the */
	/* remaining edges in place and alignment. */
	void alignEdges() {
		int aligned, color, edge;

		/* move the red side to the front */
		switch (colorSide(1, centerFace())) {
		case RIGHT_FACE:
			rotateRight();
			break;
		case BACK_FACE:
			rotateRight();
			rotateRight();
			break;
		case LEFT_FACE:
			rotateLeft();
			break;
		default:
			break;
		}

		/* rotate the top until its aligned with the center colors */
		edge = findEdge(TOP_FACE, FRONT_FACE);
		switch (edge) {
		case 2:
			rotateTopLeft();
			break;
		case 3:
			rotateTopLeft();
			rotateTopLeft();
			break;
		case 4:
			rotateTopRight();
			break;
		default:
			break;
		}

		/* rotate the bottom until its aligned with the center */
		/* colors */
		edge = findEdge(BOTTOM_FACE, FRONT_FACE);
		switch (edge) {
		case 10:
			rotateBottomLeft();
			break;
		case 11:
			rotateBottomLeft();
			rotateBottomLeft();
			break;
		case 12:
			rotateBottomRight();
			break;
		default:
			break;
		}

		if (rubik.checkSolved(false))
			return;

		rotateCcw(); /* place unaligned edges vertically */

		/* see if any edges are in correct position */
		aligned = 0;
		edge = findEdge(FRONT_FACE, RIGHT_FACE);
		if (edge == 1)
			aligned++;
		edge = findEdge(RIGHT_FACE, BACK_FACE);
		if (edge == 3)
			aligned++;
		edge = findEdge(BACK_FACE, LEFT_FACE);
		if (edge == 11)
			aligned++;
		edge = findEdge(LEFT_FACE, FRONT_FACE);
		if (edge == 9)
			aligned++;

		if (aligned == 0) {
			shuffleEdges();	/* put one edge into position */
			aligned++;
		}
		if (aligned == 1) {
			/* find the correct piece and move it to the back */
			/* bottom edge */
			edge = findEdge(FRONT_FACE, RIGHT_FACE);
			if (edge == 1) {
				rotateDown();
				rotateDown();
			} else {
				edge = findEdge(RIGHT_FACE, BACK_FACE);
				if (edge == 3)
					rotateUp();
				else {
					edge = findEdge(LEFT_FACE, FRONT_FACE);
					if (edge == 9)
						rotateDown();
				}
			}

			/* shuffle */
			color = colorSide(1, centerFace());
			if ((colorSide(1, sideFace(3)) == color) || (colorSide(6, sideFace(0)) == color)) {
				rotateRight();
				rotateRight();
				rotateDown();
				shuffleEdges();
			} else
				shuffleEdges();
		}
		/* pieces should be in place; complete color alignment */
		/* find number of unaligned edge pieces and fix them */
		aligned = 0;
		if (colorSide(1, sideFace(0)) == colorSide(1, centerFace()))
			aligned++;
		if (colorSide(1, sideFace(3)) == colorSide(1, centerFace()))
			aligned++;
		if (colorSide(3, sideFace(0)) == colorSide(3, centerFace()))
			aligned++;
		if (colorSide(3, sideFace(3)) == colorSide(3, centerFace()))
			aligned++;
		if (aligned == 0) {
			recolorTopEdges();
			rotateDown();
			rotateDown();
			recolorTopEdges();
		} else if (aligned == 2) {
			if (colorSide(1, sideFace(0)) == colorSide(1, centerFace()))
				do {
					rotateDown();
				} while (colorSide(1, sideFace(0)) == colorSide(1, centerFace()));
			if (colorSide(1, sideFace(3)) != colorSide(1, centerFace()))
				rotateUp();
			recolorTopEdges();
			if (!rubik.checkSolved(false)) {
				rotateDown();
				recolorTopEdges();
			}
		}
	}


	/* Wrong center rotation on top and front */
	void shuffle2_180() {
		for (int i = 0; i < 3; i++)
			shuffleEdges();
	}

	/* Wrong center rotation on top */
	void shuffle1_180(boolean top) {
		int i;

		if (top) {
			for (i = 0; i < 30; i++) {
				rotateTopRight();
				rotateFrontCw();
				rotateFrontCw();
			}
		} else {
			for (i = 0; i < 30; i++) {
				rotateTopRight();
				rotateTopRight();
				rotateFrontCw();
			}
		}
	}

	/* Pairs off by 90 degrees */
	void shuffle2_90(boolean counterTop, boolean counterFront) {
		int i, num;

		num = ((counterTop && !counterFront) ||
			(!counterTop && counterFront)) ? 63 : 105;
		for (i = 0; i < num; i++) {
			if (counterFront)
				rotateFrontCcw();
			else
				rotateFrontCw();
			if (counterTop)
				rotateTopRight();
			else
				rotateTopLeft();
		}
	}

	void solveAdjCenters() {
		int face0 = mapFace(5), face1 = mapFace(1);

		if ((face0 == 2 && face1 != 0) ||
				(face1 == 2 && face0 != 0)) {
			shuffle2_180();
		} else if (face0 == 2 && face1 == 0) {
			shuffle1_180(true);
		} else if (face0 == 0 && face1 == 2) {
			shuffle1_180(false);
		} else if (face0 == 1 && face1 == 1) {
			shuffle2_90(true, true);
		} else if (face0 == 3 && face1 == 1) {
			shuffle2_90(true, false);
		} else if (face0 == 1 && face1 == 3) {
			shuffle2_90(false, true);
		} else if (face0 == 3 && face1 == 3) {
			shuffle2_90(false, false);
		}
	}

	void solveOppCenters() {
		int face0, face1, face2;
		boolean fixOp = false;

		face0 = mapFace(5);
		face1 = mapFace(1);
		face2 = mapFace(4);
		if ((face1 == 1) || (face1 == 3)) {
			rotateUp();
			fixOp = true;
		} else if ((face2 == 1) || (face2 == 3)) {
			rotateCw();
			fixOp = true;
		} else if ((face0 == 1) || (face0 == 3))
			fixOp = true;
		if (!fixOp)
			return;
		face0 = mapFace(5);
		face1 = mapFace(1);
		if (face0 == 1)
			shuffle2_90(false, true);
		else /* face0 == 3 */
			shuffle2_90(true, false);
		rotateUp();
		face0 = mapFace(5);
		face1 = mapFace(1);
		if (face0 == 1 && face1 == 1) {
			shuffle2_90(true, true);
		} else if (face0 == 3 && face1 == 1) {
			shuffle2_90(true, false);
		} else if (face0 == 1 && face1 == 3) {
			shuffle2_90(false, true);
		} else if (face0 == 3 && face1 == 3) {
			shuffle2_90(false, false);
		}
	}

	void solveCenters() {
		solveAdjCenters(); /* 5-1 */
		rotateUp();
		rotateUp();
		solveAdjCenters(); /* 6-3 */
		rotateCw();
		solveAdjCenters(); /* 4-3 */
		rotateUp();
		solveAdjCenters(); /* 3-2 */
		/* only 90 degree matches left */
		rotateUp();
		solveAdjCenters(); /* 2-1 */
		rotateUp();
		solveAdjCenters(); /* 1-4 */
		rotateCcw();
		solveAdjCenters(); /* 5-4 */
		rotateUp();
		solveAdjCenters(); /* 4-6 */
		rotateUp();
		solveAdjCenters(); /* 6-2 */
		rotateUp();
		solveAdjCenters(); /* 2-5 */
		rotateCcw();
		rotateUp();
		solveAdjCenters(); /* 1-6 */
		rotateUp();
		rotateUp();
		solveAdjCenters(); /* 3-5 */
		/* only 1 90 degree match left and must be on opposite sides */
		solveOppCenters();
	}

	/* Revenge */

	void solve4USide(int side, int clockEdge) {
		int parity = clockEdge & 1;
		int edge = clockEdge >> 1;
		RubikFE rubikFE = findClockEdge(side, clockEdge);
		int foundEdge = rubikFE.edge >> 1;

		if (parity == 1)
			parity += rubik.sizeX - 4;
		if (!(side == LEFT_FACE || side == RIGHT_FACE)) {
			System.out.println("choose a side");
			return;
		}

		if (edge == 0) {
			return; /* solve this later, working on U */
		}
		if (rubikFE.face == TOP_FACE) {
			if (foundEdge == 0) {
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			} else if (foundEdge == 1) {
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopLeft();
			} else if (foundEdge == 2) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			} else if (foundEdge == 3) {
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopRight();
			}
		} else if (rubikFE.face == LEFT_FACE) {
			if (rubikFE.edge == clockEdge && side == LEFT_FACE) {
				return; /* left do nothing */
			}
			/* if (foundEdge == 0) do nothing */
			if (foundEdge == 1)
				rotateLeftUp();
			else if (foundEdge == 2) {
				rotateLeftUp();
				rotateLeftUp();
			} else if (foundEdge == 3)
				rotateLeftDown();
			rotateTopRight();
			moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			rotateTopLeft();
			/* if (foundEdge == 0) do nothing */
			if (foundEdge == 1)
				rotateLeftDown();
			else if (foundEdge == 2) {
				rotateLeftUp();
				rotateLeftUp();
			} else if (foundEdge == 3)
				rotateLeftUp();
		} else if (rubikFE.face == FRONT_FACE) {
			if (foundEdge == 0)
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			else if (foundEdge == 1) {
				rotateRightUp();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateRightDown();
			} else if (foundEdge == 2) {
				rotateFrontCw();
				rotateFrontCw();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateFrontCw();
				rotateFrontCw();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			} else if (foundEdge == 3) {
				rotateFrontCw();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateFrontCcw();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			}
		} else if (rubikFE.face == RIGHT_FACE) {
			if (rubikFE.edge == clockEdge && side == RIGHT_FACE) {
				return; /* right do nothing */
			}
			/* if (foundEdge == 0) do nothing */
			if (foundEdge == 1)
				rotateRightDown();
			else if (foundEdge == 2) {
				rotateRightUp();
				rotateRightUp();
			} else if (foundEdge == 3)
				rotateRightUp();
			rotateTopLeft();
			moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			rotateTopRight();
			/* if (foundEdge == 0) do nothing */
			if (foundEdge == 1)
				rotateRightUp();
			else if (foundEdge == 2) {
				rotateRightUp();
				rotateRightUp();
			} else if (foundEdge == 3)
				rotateRightDown();
		} else if (rubikFE.face == BOTTOM_FACE) {
			/* if (foundEdge == 0) do nothing */
			if (foundEdge == 1) {
				rotateRightUp();
				rotateRightUp();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateRightDown();
				rotateRightDown();
			} else if (foundEdge == 2) {
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
			} else if (foundEdge == 3) {
				rotateLeftUp();
				rotateLeftUp();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopRight();
				rotateLeftDown();
				rotateLeftDown();
			}
		} else if (rubikFE.face == BACK_FACE) {
			if (foundEdge == 0)
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
			else if (foundEdge == 1) {
				rotateRightDown();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateRightUp();
			} else if (foundEdge == 2) {
				rotateTopRight();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopRight();
				rotateTopRight();
			} else if (foundEdge == 3) {
				rotateLeftDown();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.BOTTOM, false);
				rotateTopRight();
				rotateLeftUp();
			}
		}
		if (side == LEFT_FACE) {
			if (edge == 1) {
				rotateLeftUp();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopLeft();
				rotateLeftDown();
			} else if (edge == 2) {
				rotateLeftUp();
				rotateLeftUp();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopLeft();
				rotateLeftDown();
				rotateLeftDown();
			} else if (edge == 3) {
				rotateLeftDown();
				rotateTopRight();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopLeft();
				rotateLeftUp();
			}
		} else if (side == RIGHT_FACE) {
			if (edge == 1) {
				rotateRightDown();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopRight();
				rotateRightUp();
			} else if (edge == 2) {
				rotateRightUp();
				rotateRightUp();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopRight();
				rotateRightDown();
				rotateRightDown();
			} else if (edge == 3) {
				rotateRightUp();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, 1 + parity, Rubik.TOP, false);
				rotateTopRight();
				rotateRightDown();
			}
		}
	}

	void solve1stTopPairEdge4(int edge) {
		int clockEdge0 = (edge << 1);
		int clockEdge1 = clockEdge0 + 1;
		RubikFE rubikFE0 = findClockEdge(TOP_FACE, clockEdge0);
		RubikFE rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		int foundEdge0 = rubikFE0.edge >> 1;
		int foundEdge1 = rubikFE1.edge >> 1;
		Boolean onSide0 = false;
		Boolean onSide1 = false;
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		if (rubikFE0.face == TOP_FACE && rubikFE1.face == TOP_FACE &&
				foundEdge0 == edge && foundEdge1 == edge) {
			return; /* done already */
		}
		/* 3 possibilities 0 on sides, 1 on side or 2 on sides */
		if (rubikFE0.face == LEFT_FACE || rubikFE0.face == RIGHT_FACE ||
				(rubikFE0.face == TOP_FACE && ((rubikFE0.edge >> 1) & 1) == 1))
			onSide0 = true;
		if (rubikFE1.face == LEFT_FACE || rubikFE1.face == RIGHT_FACE ||
				(rubikFE1.face == TOP_FACE && ((rubikFE1.edge >> 1) & 1) == 1))
			onSide1 = true;
		/* preallignment */
		if (onSide0 && onSide1) { /* moving 2 */
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopLeft();
		} else if (onSide0 || onSide1) {
			if (!(onSide0 && onSide1) &&
					(rubikFE0.face == BOTTOM_FACE || rubikFE1.face == BOTTOM_FACE ||
					(rubikFE0.face == FRONT_FACE && (rubikFE0.edge >> 1) == 2) ||
					(rubikFE1.face == FRONT_FACE && (rubikFE1.edge >> 1) == 2) ||
					(rubikFE0.face == BACK_FACE && (rubikFE0.edge >> 1) == 0) ||
					(rubikFE1.face == BACK_FACE && (rubikFE1.edge >> 1) == 0))) {
				/* moving out of way */
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			}
			if (rubikFE0.face == RIGHT_FACE || rubikFE1.face == RIGHT_FACE ||
					(rubikFE0.face == Rubik.TOP && (rubikFE0.edge >> 1) == 1) ||
					(rubikFE1.face == Rubik.TOP && (rubikFE1.edge >> 1) == 1)) {
				/* moving 1 front TR */
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
				rotateFrontCw();
				rotateFrontCw();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateFrontCw();
				rotateFrontCw();
				rotateTopRight();
			} else {
				/* moving 1 front TL */
				rotateTopRight();
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
				rotateFrontCw();
				rotateFrontCw();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateFrontCw();
				rotateFrontCw();
				rotateTopLeft();
			}
		}
		if (onSide0 || onSide1) {
			rubikFE0 = findClockEdge(TOP_FACE, clockEdge0);
			rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		}
		if (!(rubikFE0.face == BOTTOM_FACE && rubikFE0.edge == 0)) {
			if (rubikFE0.face == TOP_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == FRONT_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == BACK_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			} else if (rubikFE0.face == FRONT_FACE && rubikFE0.edge == 4) {
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == TOP_FACE && rubikFE0.edge == 4) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			} else if (rubikFE0.face == BOTTOM_FACE && rubikFE0.edge == 4) {
				rotateBackCw();
				rotateBackCw();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateBackCw();
				rotateBackCw();
			} else if (rubikFE0.face == BACK_FACE && rubikFE0.edge == 4) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			}
			rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		}
		if (!(rubikFE1.face == BOTTOM_FACE && rubikFE1.edge == 1)) {
			if (rubikFE1.face == TOP_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			} else if (rubikFE1.face == FRONT_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			} else if (rubikFE1.face == BACK_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			} else if (rubikFE1.face == FRONT_FACE && rubikFE1.edge == 5) {
				System.out.println("impossible 2 5 " + rubikFE1.face + " " + rubikFE1.edge);
			} else if (rubikFE1.face == TOP_FACE && rubikFE1.edge == 5) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			} else if (rubikFE1.face == BOTTOM_FACE && rubikFE1.edge == 5) {
				rotateBackCw();
				rotateBackCw();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateBackCw();
				rotateBackCw();
			} else if (rubikFE1.face == BACK_FACE && rubikFE1.edge == 5) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			}
		}
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
	}

	void solve2ndTopPairEdge4(int edge) {
		int clockEdge0 = (edge << 1);
		int clockEdge1 = clockEdge0 + 1;
		RubikFE rubikFE0 = findClockEdge(TOP_FACE, clockEdge0);
		RubikFE rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		int foundEdge0 = rubikFE0.edge >> 1;
		int foundEdge1 = rubikFE1.edge >> 1;
		Boolean onSide0 = false;
		Boolean onSide1 = false;
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		if (rubikFE0.face == TOP_FACE && rubikFE1.face == TOP_FACE &&
				foundEdge0 == edge && foundEdge1 == edge) {
			/* done already */
			return;
		}
		/* 3 possibilities 0 on sides, 1 on side or 2 on sides */
		if (rubikFE0.face == LEFT_FACE || /* rubikFE0.face == RIGHT_FACE || */
				(rubikFE0.face == TOP_FACE && ((rubikFE0.edge >> 1) & 1) == 1))
			onSide0 = true;
		if (rubikFE1.face == LEFT_FACE || /* rubikFE1.face == RIGHT_FACE || */
				(rubikFE1.face == TOP_FACE && ((rubikFE1.edge >> 1) & 1) == 1))
			onSide1 = true;
		/* preallignment */
		if (onSide0 || onSide1) {
			if (!(onSide0 && onSide1) &&
					(rubikFE0.face == BOTTOM_FACE || rubikFE1.face == BOTTOM_FACE ||
					(rubikFE0.face == FRONT_FACE && (rubikFE0.edge >> 1) == 2) ||
					(rubikFE1.face == FRONT_FACE && (rubikFE1.edge >> 1) == 2) ||
					(rubikFE0.face == BACK_FACE && (rubikFE0.edge >> 1) == 0) ||
					(rubikFE1.face == BACK_FACE && (rubikFE1.edge >> 1) == 0))) {
				/* moving out of way */
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			}
			/* moving 1 front TR */
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateFrontCw();
			rotateFrontCw();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateFrontCw();
			rotateFrontCw();
			rotateTopLeft();
		}
		if (onSide0 || onSide1) {
			rubikFE0 = findClockEdge(TOP_FACE, clockEdge0);
			rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		}
		if (!(rubikFE0.face == BOTTOM_FACE && rubikFE0.edge == 0)) {
			if (rubikFE0.face == TOP_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == FRONT_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == BACK_FACE && rubikFE0.edge == 0) {
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			} else if (rubikFE0.face == FRONT_FACE && rubikFE0.edge == 4) {
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				rotateBottomLeft();
				rotateBottomLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			} else if (rubikFE0.face == TOP_FACE && rubikFE0.edge == 4) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			} else if (rubikFE0.face == BOTTOM_FACE && rubikFE0.edge == 4) {
				rotateBackCw();
				rotateBackCw();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateBackCw();
				rotateBackCw();
			} else if (rubikFE0.face == BACK_FACE && rubikFE0.edge == 4) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			}
			rubikFE1 = findClockEdge(TOP_FACE, clockEdge1);
		}
		if (!(rubikFE1.face == BOTTOM_FACE && rubikFE1.edge == 1)) {
			if (rubikFE1.face == TOP_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			} else if (rubikFE1.face == FRONT_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			} else if (rubikFE1.face == BACK_FACE && rubikFE1.edge == 1) {
				moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			} else if (rubikFE1.face == FRONT_FACE && rubikFE1.edge == 5) {
				System.out.println("impossible 2 5 " + rubikFE1.face + " " + rubikFE1.edge);
			} else if (rubikFE1.face == TOP_FACE && rubikFE1.edge == 5) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			} else if (rubikFE1.face == BOTTOM_FACE && rubikFE1.edge == 5) {
				rotateBackCw();
				rotateBackCw();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateBackCw();
				rotateBackCw();
			} else if (rubikFE1.face == BACK_FACE && rubikFE1.edge == 5) {
				rotateTopLeft();
				rotateTopLeft();
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
				rotateTopLeft();
				rotateTopLeft();
			}
		}
		rotateTopRight();
		rotateFrontCw();
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopLeft();
	}

	void mix3Edges() {
		int pair1 = rubik.sizeX - 2;
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		rotateTopRight();
	}

	void flipDiagonal2Pairs() {
		int pair1 = rubik.sizeX - 2;

		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
	}

	void rotate1stPair(int pair, int face) {
		switch (face) {
		case 1:
			moveAPiece(FRONT_FACE, pair, Rubik.BOTTOM, false);
			return;
		case 2:
			moveAPiece(FRONT_FACE, pair, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair, Rubik.BOTTOM, false);
			return;
		case 3:
			moveAPiece(FRONT_FACE, pair, Rubik.TOP, false);
			return;
		default:
			return;
		}
	}

	void rotate2ndPair(int face) {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;
		switch (face) {
		case 2:
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			return;
		case 1:
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateTopRight();
			rotateTopRight();
			return;
		default:
			return;
		}
	}

	void topPairFlip() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		rotateTopRight();
	}

	void frontPairFlip() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateFrontCw();
		rotateFrontCw();
	}

	boolean pair1stEdges() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;
		int face0, face1;

		for (face0 = 0; face0 < 4; face0++) {
			for (face1 = 0; face1 < 4; face1++) {
				if (matchEdges(lastColumn[face0], lastColumn[face1], 0)) {
					rotate1stPair(pair0, face0);
					rotate1stPair(pair1, face1);
					return true;
				}
			}
		}
		return false;
	}

	boolean pair2ndEdges() {
		int face;

		for (face = 1; face < 4; face++) {
			int last = lastColumn[face];
			if (matchEdges(last, last, 0)) {
				rotate2ndPair(face);
				return true;
			}
		}
		return false;
	}

	void pair1stPairOf8Edges() {
		for (int i = 0; i < 2; i++) {
			if (pair1stEdges()) {
				return;
			} else {
				mix3Edges();
			}
		}
		/* 1st not working */
	}

	void pair2ndPairOf8Edges() {
		for (int i = 0; i < 5; i++) {
			if (pair2ndEdges()) {
				return;
			} else {
				mix3Edges();
			}
		}
		/* 2nd not working */
	}

	void pairLastEdges() {
		int quad0 = 1;
		int quad1 = rubik.sizeX - 2;
		int quad2 = rubik.sizeX * rubik.sizeX - rubik.sizeX + 1;
		int quad3 = rubik.sizeX * rubik.sizeX - 2;

		if (rubik.cubeLoc[FRONT_FACE][quad0].face ==
				rubik.cubeLoc[FRONT_FACE][quad2].face &&
				rubik.cubeLoc[FRONT_FACE][quad1].face ==
				rubik.cubeLoc[FRONT_FACE][quad3].face) {
			topPairFlip();
		}
		if (!(rubik.cubeLoc[FRONT_FACE][quad0].face ==
				rubik.cubeLoc[FRONT_FACE][quad1].face &&
				rubik.cubeLoc[FRONT_FACE][quad2].face ==
				rubik.cubeLoc[FRONT_FACE][quad3].face)) {
			flipDiagonal2Pairs();
		}
	}

	void solve42U() {
		solve4USide(LEFT_FACE, 2);
		solve4USide(LEFT_FACE, 3);
		solve4USide(LEFT_FACE, 4);
		solve4USide(LEFT_FACE, 5);
		solve4USide(LEFT_FACE, 6);
		solve4USide(LEFT_FACE, 7);
		solve4USide(RIGHT_FACE, 2);
		solve4USide(RIGHT_FACE, 3);
		solve4USide(RIGHT_FACE, 4);
		solve4USide(RIGHT_FACE, 5);
		solve4USide(RIGHT_FACE, 6);
		solve4USide(RIGHT_FACE, 7);
	}

	void solve2PairEdge4() {
		solve1stTopPairEdge4(1);
		solve2ndTopPairEdge4(3);
	}

	void pairLast8Edges() {
		pair1stPairOf8Edges();
		pair2ndPairOf8Edges();
		pairLastEdges();
	}


	void place4thPair() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;
		RubikFE rubikFE = findClockEdge(BACK_FACE, 0);
		/* if (rubikFE.face == BACK_FACE && rubikFE.edge == 0)
			return; *//* 4th pair already set */
		if (rubikFE.face == TOP_FACE && rubikFE.edge == 0) {
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		} else if (rubikFE.face == FRONT_FACE && rubikFE.edge == 0) {
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		} else if (rubikFE.face == BOTTOM_FACE && rubikFE.edge == 0) {
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		} else if (rubikFE.face == BACK_FACE && rubikFE.edge == 4) {
			rotateBackCw();
			rotateBackCw();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateBackCw();
			rotateBackCw();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		} else if (rubikFE.face == TOP_FACE && rubikFE.edge == 4) {
			rotateBackCw();
			rotateBackCw();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateBackCw();
			rotateBackCw();
		} else if (rubikFE.face == FRONT_FACE && rubikFE.edge == 4) {
			rotateFrontCw();
			rotateFrontCw();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateFrontCw();
			rotateFrontCw();
		} else if (rubikFE.face == BOTTOM_FACE && rubikFE.edge == 4) {
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateBottomRight();
			rotateBottomRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateBottomRight();
			rotateBottomRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		}
	}

	/*void topPairExchange() {
		int pair1 = rubik.sizeX - 2;

		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
	}*/

	void frontPairExchange() {
		int pair1 = rubik.sizeX - 2;

		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateFrontCw();
		rotateFrontCw();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
	}

	void place3rdPair() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;
		RubikFE rubikFE = findClockEdge(TOP_FACE, 0);
		/*if (rubikFE.face == TOP_FACE && rubikFE.edge == 0)
			return; *//* 3rd pair already set */
		if (rubikFE.face == FRONT_FACE && rubikFE.edge == 0) {
			topPairFlip();
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		} else if (rubikFE.face == BOTTOM_FACE && rubikFE.edge == 0) {
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateTopRight();
			rotateTopRight();
		} else if (rubikFE.face == BACK_FACE && rubikFE.edge == 4) {
			topPairFlip();
		} else if (rubikFE.face == TOP_FACE && rubikFE.edge == 4) {
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		} else if (rubikFE.face == FRONT_FACE && rubikFE.edge == 4) {
			frontPairFlip();
			moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateTopRight();
			rotateTopRight();
		}
	}

	void upsetSingleTopFrontPair() {
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		/* upset */
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		rotateBottomLeft();
		rotateBottomLeft();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		rotateBottomLeft();
		rotateBottomLeft();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateTopRight();
		rotateTopRight();
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);

		/* vb */
		moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
		rotateBackCw();
		rotateBackCw();
		moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
		moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
		rotateBackCw();
		rotateBackCw();

		/* flip  */
		rotateFrontCw();
		rotateFrontCw();
		topPairFlip();
		rotateFrontCw();
		rotateFrontCw();
	}

	void place2ndPair() {
		RubikFE rubikFE0 = findClockEdge(FRONT_FACE, 0);
		RubikFE rubikFE1 = findClockEdge(FRONT_FACE, 4);
		int pair0 = 1;
		int pair1 = rubik.sizeX - 2;

		if ((rubikFE0.face == FRONT_FACE && rubikFE0.edge == 4) ||
				(rubikFE1.face == FRONT_FACE && rubikFE1.edge == 0)) {
			frontPairExchange();
		} else if (rubikFE0.face != FRONT_FACE && rubikFE1.face != FRONT_FACE &&
				(rubikFE0.edge == 4 || rubikFE1.edge == 0)) {
			frontPairFlip();
		} else if ((rubikFE0.face == BOTTOM_FACE && rubikFE0.edge == 0) ||
				(rubikFE1.face == TOP_FACE && rubikFE1.edge == 4)) {
			frontPairExchange();
			frontPairFlip();
		}
		rubikFE0 = findClockEdge(FRONT_FACE, 0);
		rubikFE1 = findClockEdge(FRONT_FACE, 4);
		if (rubikFE0.face == TOP_FACE && rubikFE0.edge == 4) {
			if (rubikFE1.face == FRONT_FACE) {
				upsetSingleTopFrontPair();
			}
		} else if (rubikFE1.face == BOTTOM_FACE && rubikFE1.edge == 0) {
			if (rubikFE0.face == FRONT_FACE) {
				/* prepare upset */
				moveAPiece(FRONT_FACE, pair0, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
				upsetSingleTopFrontPair();
				moveAPiece(FRONT_FACE, pair0, Rubik.BOTTOM, false);
				moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			}
		}
	}

	void place4EdgePairs() {
		place4thPair();
		place3rdPair();
		place2ndPair();
	}

	void solveLast8Edges() {
		pairLast8Edges();
		place4EdgePairs();
	}

	/*void adj2PairFrontTop() {
		int pair1 = rubik.sizeX - 2;
		int i;

		for (i = 0; i < 5; i++) {
			moveAPiece(FRONT_FACE, pair1, Rubik.TOP, false);
			rotateTopRight();
			rotateTopRight();
			moveAPiece(FRONT_FACE, pair1, Rubik.BOTTOM, false);
			rotateTopRight();
			rotateTopRight();
		}
	}*/

	void rotateFace(int face, int rotation) {
		int rot = (rotation + RubikInterface.MAX_ORIENT
			* RubikInterface.MAX_ORIENT) % RubikInterface.MAX_ORIENT;
		if (rot == 0)
			return;
		if (face == TOP_FACE) {
			if (rot == 1) {
				rotateTopLeft();
			} else if (rot == 2) {
				rotateTopLeft();
				rotateTopLeft();
			} else if (rot == 3) {
				rotateTopRight();
			}
		} else if (face == LEFT_FACE) {
			if (rot == 1) {
				rotateLeftDown();
			} else if (rot == 2) {
				rotateLeftDown();
				rotateLeftDown();
			} else if (rot == 3) {
				rotateLeftUp();
			}
		} else if (face == FRONT_FACE) {
			if (rot == 1) {
				rotateFrontCw();
			} else if (rot == 2) {
				rotateFrontCw();
				rotateFrontCw();
			} else if (rot == 3) {
				rotateFrontCcw();
			}
		} else if (face == RIGHT_FACE) {
			if (rot == 1) {
				rotateRightUp();
			} else if (rot == 2) {
				rotateRightUp();
				rotateRightUp();
			} else if (rot == 3) {
				rotateRightDown();
			}
		} else if (face == BOTTOM_FACE) {
			if (rot == 1) {
				rotateBottomRight();
			} else if (rot == 2) {
				rotateBottomRight();
				rotateBottomRight();
			} else if (rot == 3) {
				rotateBottomLeft();
			}
		} else if (face == BACK_FACE) {
			if (rot == 1) {
				rotateBackCw();
			} else if (rot == 2) {
				rotateBackCw();
				rotateBackCw();
			} else if (rot == 3) {
				rotateBackCcw();
			}
		}
	}

	void opp2PairMove(int fromFace, int toFace) {
		int quad = rubik.sizeX + 1;

		if (fromFace == TOP_FACE && toFace == BOTTOM_FACE) {
			for (int i = 0; i < 2; i++) {
				moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
				moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
				moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
				moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			}
		} else if (fromFace == LEFT_FACE && toFace == RIGHT_FACE) {
			for (int i = 0; i < 2; i++) {
				moveAPiece(LEFT_FACE, quad, Rubik.RIGHT, false);
				moveAPiece(LEFT_FACE, quad, Rubik.RIGHT, false);
				moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
				moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			}
		} else if (fromFace == FRONT_FACE && toFace == BACK_FACE) {
			for (int i = 0; i < 2; i++) {
				moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
				moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
				moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			}
		} else {
			System.out.println("op2PairMove unknown " + fromFace + " " + toFace);
		}
	}

	int bestEmptySpot(int face, int fromFace, int offset) {
		int center, position;

		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i] + offset) % RubikInterface.MAX_ORIENT;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[face][position].face == fromFace ||
				rubik.cubeLoc[face][position].face == oppositeFace[face]) {
				return (center + RubikInterface.MAX_ORIENT - offset) % RubikInterface.MAX_ORIENT;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i] + offset) % RubikInterface.MAX_ORIENT;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[face][position].face != face &&
					rubik.cubeLoc[face][position].face != oppositeFace[fromFace]) {
				return (center + RubikInterface.MAX_ORIENT - offset) % RubikInterface.MAX_ORIENT;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i] + offset) % RubikInterface.MAX_ORIENT;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[face][position].face != face) {
				return (center + RubikInterface.MAX_ORIENT - offset) % RubikInterface.MAX_ORIENT;
			}
		}
		return -1;
	}


	int bestEmptySpotRev(int face, int fromFace, int offset) {
		int center, position;

		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i] + offset) % RubikInterface.MAX_ORIENT;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[face][position].face == fromFace ||
					rubik.cubeLoc[face][position].face == oppositeFace[face]) {
				return (center + RubikInterface.MAX_ORIENT - offset) % RubikInterface.MAX_ORIENT;
			}
		}
		return -1;
	}

	int findMoveSpot(int currentFace, int spotFace) {
		int center, position;

		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[currentFace][position].face == spotFace) {
				return center;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[currentFace][position].face == oppositeFace[currentFace]) {
				return center;
			}
		}
		return -1;
	}

	int findMoveSpotRev(int currentFace, int spotFace) {
		int center, position;

		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[currentFace][position].face != currentFace &&
					rubik.cubeLoc[currentFace][position].face != oppositeFace[spotFace]) {
				return center;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[currentFace][position].face != currentFace) {
				return center;
			}
		}
		return -1;
	}


	int bestEmptySpotsForOpp(int face, int offset) {
		int center, position;
		int otherCenter, otherPosition;

		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			otherCenter = (bestOrder[i] + 3 + offset) % 4;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			otherPosition = clockCenterToPosition[otherCenter];
			otherPosition = offsetRevenge(otherPosition); /* non revenge */
			if (rubik.cubeLoc[face][position].face == oppositeFace[face] &&
					rubik.cubeLoc[face][otherPosition].face == oppositeFace[face]) {
				return center;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i]) % 4;
			otherCenter = (bestOrder[i] + 3 + offset) % 4;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			otherPosition = clockCenterToPosition[otherCenter];
			otherPosition = offsetRevenge(otherPosition); /* non revenge */
			if ((rubik.cubeLoc[face][position].face == oppositeFace[face] ||
					rubik.cubeLoc[face][otherPosition].face == oppositeFace[face]) &&
					rubik.cubeLoc[face][position].face != face &&
					rubik.cubeLoc[face][otherPosition].face != face) {
				return center;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = (bestOrder[i]) % 4;
			otherCenter = (bestOrder[i] + 3) % 4;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			otherPosition = clockCenterToPosition[otherCenter];
			otherPosition = offsetRevenge(otherPosition); /* non revenge */
			if (rubik.cubeLoc[face][position].face != face &&
					rubik.cubeLoc[face][otherPosition].face != face) {
				return center;
			}
		}
		return -1;
	}

	int findMoveSpotsForOpp(int currentFace, int offset) {
		int center, position;
		int otherCenter, otherPosition;

		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			otherCenter = (bestOrder[i] + 3 + offset) % 4;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			otherPosition = clockCenterToPosition[otherCenter];
			otherPosition = offsetRevenge(otherPosition); /* non revenge */
			if (rubik.cubeLoc[currentFace][position].face == oppositeFace[currentFace] &&
					rubik.cubeLoc[currentFace][otherPosition].face == oppositeFace[currentFace]) {
				return center;
			}
		}
		for (int i = 0; i < 4; i++) {
			center = bestOrder[i];
			otherCenter = (bestOrder[i] + 3 + offset) % 4;
			position = clockCenterToPosition[center];
			position = offsetRevenge(position); /* non revenge */
			otherPosition = clockCenterToPosition[otherCenter];
			otherPosition = offsetRevenge(otherPosition); /* non revenge */
			if ((rubik.cubeLoc[currentFace][position].face == oppositeFace[currentFace] ||
					rubik.cubeLoc[currentFace][otherPosition].face == oppositeFace[currentFace]) &&
					rubik.cubeLoc[currentFace][position].face != currentFace &&
					rubik.cubeLoc[currentFace][otherPosition].face != currentFace) {
				return center;
			}
		}
		return -1;
	}

	void adj3Move(int fromFace, int toFace) {
		int quad = rubik.sizeX + 1;

		if (fromFace == FRONT_FACE && toFace == LEFT_FACE) {
			rotateLeftDown();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
		} else if (fromFace == LEFT_FACE && toFace == TOP_FACE) {
			rotateTopRight();
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			rotateTopLeft();
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
		} else if (fromFace == TOP_FACE && toFace == RIGHT_FACE) {
			rotateRightDown();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			rotateRightUp();
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
		} else if (fromFace == RIGHT_FACE && toFace == BACK_FACE) {
			rotateBackCcw();
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			rotateBackCw();
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
		} else if (fromFace == BACK_FACE && toFace == BOTTOM_FACE) {
			rotateBottomLeft();
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateBottomRight();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
		} else if (fromFace == BOTTOM_FACE && toFace == FRONT_FACE) {
			rotateFrontCcw();
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateFrontCw();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
		} else if (fromFace == FRONT_FACE && toFace == TOP_FACE) {
			rotateTopRight();
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(TOP_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
			rotateTopLeft();
			moveAPiece(TOP_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
		} else if (fromFace == TOP_FACE && toFace == BACK_FACE) {
			rotateBackCcw();
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			rotateBackCw();
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
		} else if (fromFace == BACK_FACE && toFace == LEFT_FACE) {
			rotateLeftDown();
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			rotateLeftUp();
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
		} else if (fromFace == LEFT_FACE && toFace == BOTTOM_FACE) {
			rotateBottomRight();
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			rotateBottomLeft();
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
		} else if (fromFace == BOTTOM_FACE && toFace == RIGHT_FACE) {
			rotateRightDown();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			rotateRightUp();
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
		} else if (fromFace == RIGHT_FACE && toFace == FRONT_FACE) {
			rotateFrontCw();
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			rotateFrontCcw();
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
		} else {
			System.out.println("adj3Move Unknown " + fromFace + " " + toFace);
		}
	}

	void adj3MoveRev(int fromFace, int toFace) {
		int quad = rubik.sizeX + 1;

		if (fromFace == FRONT_FACE && toFace == LEFT_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			rotateLeftDown();
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateLeftUp();
		} else if (fromFace == LEFT_FACE && toFace == TOP_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			rotateTopRight();
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			rotateTopLeft();
		} else if (fromFace == TOP_FACE && toFace == RIGHT_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			rotateRightDown();
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateRightUp();
		} else if (fromFace == RIGHT_FACE && toFace == BACK_FACE) {
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			rotateBackCcw();
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			rotateBackCw();
		} else if (fromFace == BACK_FACE && toFace == BOTTOM_FACE) {
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateBottomLeft();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			rotateBottomRight();
		} else if (fromFace == BOTTOM_FACE && toFace == FRONT_FACE) {
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateFrontCcw();
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			rotateFrontCw();
		} else if (fromFace == FRONT_FACE && toFace == TOP_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(TOP_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
			rotateTopRight();
			moveAPiece(TOP_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(TOP_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			rotateTopLeft();
		} else if (fromFace == TOP_FACE && toFace == BACK_FACE) {
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			rotateBackCcw();
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(TOP_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			moveAPiece(TOP_FACE, quad, Rubik.RIGHT, false);
			rotateBackCw();
		} else if (fromFace == BACK_FACE && toFace == LEFT_FACE) {
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			rotateLeftDown();
			moveAPiece(BACK_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(BACK_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BACK_FACE, quad, Rubik.LEFT, false);
			moveAPiece(BACK_FACE, quad, Rubik.TOP, false);
			rotateLeftUp();
		} else if (fromFace == LEFT_FACE && toFace == BOTTOM_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			rotateBottomRight();
			moveAPiece(LEFT_FACE, quad, Rubik.TOP, false);
			moveAPiece(FRONT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(LEFT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(FRONT_FACE, quad, Rubik.LEFT, false);
			rotateBottomLeft();
		} else if (fromFace == BOTTOM_FACE && toFace == RIGHT_FACE) {
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			rotateRightDown();
			moveAPiece(BOTTOM_FACE, quad, Rubik.LEFT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(BOTTOM_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(FRONT_FACE, quad, Rubik.TOP, false);
			rotateRightUp();
		} else if (fromFace == RIGHT_FACE && toFace == FRONT_FACE) {
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			rotateFrontCw();
			moveAPiece(RIGHT_FACE, quad, Rubik.RIGHT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.BOTTOM, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.LEFT, false);
			moveAPiece(RIGHT_FACE, quad, Rubik.TOP, false);
			rotateFrontCcw();
		} else {
			System.out.println("adj3MoveRev Unknown " + fromFace + " " + toFace);
		}
	}

	void checkForOppMove(int fromFace, int toFace, int offset) {
		for (int i = 0; i < 2; i++) {
			int foundMoveSpot = findMoveSpotsForOpp(fromFace, offset);
			int foundEmptySpot = bestEmptySpotsForOpp(toFace, offset);
			if (foundMoveSpot == -1) {
				break;
			}
			if (foundEmptySpot == -1) {
				return;
			}
			rotateFace(toFace, -foundEmptySpot);
			rotateFace(fromFace, -foundMoveSpot);
			opp2PairMove(fromFace, toFace);
			rotateFace(fromFace, foundMoveSpot);
			rotateFace(toFace, foundEmptySpot);
		}
		for (int i = 0; i < 2; i++) {
			int foundMoveSpot = findMoveSpotsForOpp(toFace, offset);
			int foundEmptySpot = bestEmptySpotsForOpp(fromFace, offset);
			if (foundMoveSpot == -1 || foundEmptySpot == -1) {
				return;
			}
			rotateFace(fromFace, -foundEmptySpot);
			rotateFace(toFace, -foundMoveSpot);
			opp2PairMove(fromFace, toFace);
			rotateFace(toFace, foundMoveSpot);
			rotateFace(fromFace, foundEmptySpot);
		}
	}

	void checkForAdjMove(int fromFace, int toFace, int offset) {
		for (int i = 0; i < 4; i++) {
			int foundMoveSpot = findMoveSpot(fromFace, toFace);
			int foundEmptySpot = bestEmptySpot(toFace, fromFace, offset);
			int movePosition, emptyPosition;
			if (foundMoveSpot == -1) {
				break;
			}
			if (foundEmptySpot == -1) {
				return;
			}
			movePosition = clockCenterToPosition[foundMoveSpot];
			emptyPosition = clockCenterToPosition[(foundEmptySpot + offset) % RubikInterface.MAX_ORIENT];
			movePosition = offsetRevenge(movePosition); /* non revenge */
			emptyPosition = offsetRevenge(emptyPosition); /* non revenge */
			if (rubik.cubeLoc[fromFace][movePosition].face ==
					rubik.cubeLoc[toFace][emptyPosition].face) {
				/* do not make move of same color */
				break;
			}
			rotateFace(toFace, -foundEmptySpot);
			rotateFace(fromFace, -foundMoveSpot);
			adj3Move(fromFace, toFace);
			rotateFace(fromFace, foundMoveSpot);
			rotateFace(toFace, foundEmptySpot);
		}
		for (int i = 0; i < 4; i++) {
			int foundMoveSpot = findMoveSpotRev(fromFace, toFace);
			int foundEmptySpot = bestEmptySpotRev(toFace, fromFace, offset);
			int movePosition, emptyPosition;
			if (foundMoveSpot == -1 || foundEmptySpot == -1) {
				return;
			}
			movePosition = clockCenterToPosition[foundMoveSpot];
			emptyPosition = clockCenterToPosition[(foundEmptySpot + offset) % RubikInterface.MAX_ORIENT];
			movePosition = offsetRevenge(movePosition); /* non revenge */
			emptyPosition = offsetRevenge(emptyPosition); /* non revenge */
			if (rubik.cubeLoc[fromFace][movePosition].face ==
					rubik.cubeLoc[toFace][emptyPosition].face) {
				/* do not make move of same color */
				return;
			}
			rotateFace(toFace, -foundEmptySpot);
			rotateFace(fromFace, -foundMoveSpot);
			adj3Move(fromFace, toFace);
			rotateFace(fromFace, foundMoveSpot);
			rotateFace(toFace, foundEmptySpot);
		}
	}

	void mixUp(int face, int clockEdge) {
		int position = clockCenterToPosition[clockEdge];
		position = offsetRevenge(position); /* non revenge */

		moveAPiece(face, position, Rubik.LEFT, false);
		moveAPiece(face, position, Rubik.BOTTOM, false);
		moveAPiece(face, position, Rubik.RIGHT, false);
		moveAPiece(face, position, Rubik.TOP, false);
	}

	RubikFE centersSolvedOnFace(int face) {
		RubikFE rubikFE = new RubikFE();

		int color = rubik.cubeLoc[face][0].face;
		for (int i = 0; i < RubikInterface.MAX_ORIENT; i++) {
			int position = clockCenterToPosition[i];
			position = offsetRevenge(position); /* non revenge */
			if (rubik.cubeLoc[face][position].face != color) {
				rubikFE.face = face;
				rubikFE.edge = i;
				return rubikFE;
			}
		}
		rubikFE.face = -1;
		rubikFE.edge = -1;
		return rubikFE;
	}


	RubikFE centersSolved() {
		RubikFE rubikFE = new RubikFE();
		rubikFE.face = -1;
		rubikFE.edge = -1;

		for (int face = 0; face < Rubik.MAX_FACES; face++) {
			rubikFE = centersSolvedOnFace(face);
			if (rubikFE.face != -1) {
				return rubikFE;
			}
		}
		return rubikFE;
	}

	void solveAdj() {
		for (int i = 0; i < 2; i++) {
			checkForAdjMove(FRONT_FACE, LEFT_FACE, 0);
			checkForAdjMove(LEFT_FACE, TOP_FACE, 1);
			checkForAdjMove(TOP_FACE, RIGHT_FACE, 1);
			checkForAdjMove(RIGHT_FACE, BACK_FACE, 2);
			checkForAdjMove(BACK_FACE, BOTTOM_FACE, 0);
			checkForAdjMove(BOTTOM_FACE, FRONT_FACE, 0);

			checkForAdjMove(FRONT_FACE, TOP_FACE, 0);
			checkForAdjMove(TOP_FACE, BACK_FACE, 0);
			checkForAdjMove(BACK_FACE, LEFT_FACE, 2);
			checkForAdjMove(LEFT_FACE, BOTTOM_FACE, 3);
			checkForAdjMove(BOTTOM_FACE, RIGHT_FACE, 3);
			checkForAdjMove(RIGHT_FACE, FRONT_FACE, 0);
		}
	}

	void solveCenter4() {
		RubikFE rubikFE;

		checkForOppMove(TOP_FACE, BOTTOM_FACE, 0);
		checkForOppMove(FRONT_FACE, BACK_FACE, 0);
		checkForOppMove(LEFT_FACE, RIGHT_FACE, 2);
		solveAdj();
		/* rare case of opposites unsolved */
		rubikFE = centersSolved();
		if (rubikFE.face >= 0) {
			/* trying again */
			mixUp(rubikFE.face, rubikFE.edge);
			rubikFE.face = oppositeFace[rubikFE.face];
			rubikFE = centersSolvedOnFace(rubikFE.face);
			mixUp(rubikFE.face, rubikFE.edge);
			solveAdj();
		}
	}

	void
	orientCentersFace(int face, int adjFace, int faceRotation)
	{
		int i, clockCenter, orient;
		int faceColor = rubik.cubeLoc[face][0].face;
		/*int faceRotation = rubik.cubeLoc[face][0].rotation;
		faceRotation = 0;*/
		/* for each piece out of place */
		for (i = 0; i < 4; i++) {
			clockCenter = clockCenterToPosition[i];
			orient = mapFaceOrientRev(face, clockCenter);
			if (orient != 0) {
				if (i != 0) {
					rotateFace(face, RubikInterface.MAX_ORIENT - i);
				}
				/* rotate, take out using adj3MoveRev */
				adj3MoveRev(face, adjFace);
				/* then rotate put back and undo with adj3Move */
				rotateFace(face, orient);
				adj3Move(face, adjFace);
				rotateFace(face, RubikInterface.MAX_ORIENT - orient);
				clockCenter = clockCenterToPosition[3];
				orient = mapFaceOrientRev(adjFace, clockCenter);
				rotateFace(adjFace, 1);
				rotateFace(face, orient + 1 + i + faceRotation);
				adj3Move(face, adjFace);
				rotateFace(face, RubikInterface.MAX_ORIENT - orient - 1 - i - faceRotation);
				rotateFace(adjFace, 3);
	
				clockCenter = clockCenterToPosition[2];
				if (faceColor != rubik.cubeLoc[adjFace][clockCenter].face) {
					if (i != 0) {
						rotateFace(face, i);
					}
					continue;
				}
				orient = mapFaceOrientRev(adjFace, clockCenter);
				rotateFace(adjFace, 2);
				rotateFace(face, orient + 2 + i + faceRotation);
				adj3Move(face, adjFace);
				rotateFace(face, RubikInterface.MAX_ORIENT - orient - 2 - i - faceRotation);
				rotateFace(adjFace, 2);
				clockCenter = clockCenterToPosition[1];
				if (faceColor != rubik.cubeLoc[adjFace][clockCenter].face) {
					if (i != 0) {
						rotateFace(face, i);
					}
					continue;
				}
				orient = mapFaceOrientRev(adjFace, clockCenter);
				rotateFace(adjFace, 3);
				rotateFace(face, orient + 3 + i + faceRotation);
				adj3Move(face, adjFace);
				rotateFace(face, RubikInterface.MAX_ORIENT - orient - 3 - i - faceRotation);
				rotateFace(adjFace, 1);
				if (i != 0) {
					rotateFace(face, i);
				}
			}
		}
	}
	
	void orientLastCenterFace() {
		/* 12 possibilities
		  1 solved, 8 posssiblities with 3 cubies are wrong,
		  3 possibilities with all 4 cubies are wrong
		  last 3 are solved by moving 3 cubies twice. */
		int clockCenter0 = clockCenterToPosition[0];
		int clockCenter1 = clockCenterToPosition[1];
		int rotation0 = rubik.cubeLoc[2][clockCenter0].rotation;
		int rotation1 = rubik.cubeLoc[2][clockCenter1].rotation;
		int quad2 = rubik.sizeX * rubik.sizeX - rubik.sizeX - 1;

		if (rotation0 == 0 && rotation1 == 2) { /* 3 off 0211 */
			rotateFrontCw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3MoveRev(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCcw();
		} else if (rotation0 == 0 && rotation1 == 3) { /* 3 off 0332 */
			rotateFrontCw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3Move(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCcw();
		} else if (rotation0 == 1 && rotation1 == 0) { /* 3 off 1021 */
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3MoveRev(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
		} else if (rotation0 == 1 && rotation1 == 1) { /* 3 off 1102 */
			rotateFrontCcw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3MoveRev(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCw();
		} else if (rotation0 == 2 && rotation1 == 0) { /* 3 off 2033 */
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3Move(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
		} else if (rotation0 == 2 && rotation1 == 1) { /* 3 off 2110 */
			rotateFrontCw();
			rotateFrontCw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3MoveRev(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCcw();
			rotateFrontCcw();
		} else if (rotation0 == 3 && rotation1 == 2) { /* 3 off 3203 */
			rotateFrontCcw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3Move(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCw();
		} else if (rotation0 == 3 && rotation1 == 3) { /* 3 off 3320 */
			rotateFrontCw();
			rotateFrontCw();
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3Move(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			rotateFrontCcw();
			rotateFrontCcw();
		} else if ((rotation0 == 1 && rotation1 == 3) /* 4 off 1331 */
				|| (rotation0 == 2 && rotation1 == 2) /* 4 off 2222 */
				|| (rotation0 == 3 && rotation1 == 1)) { /* 4 off 3113 */
			moveAPiece(FRONT_FACE, quad2, Rubik.LEFT, false);
			rotateLeftDown();
			adj3Move(FRONT_FACE, LEFT_FACE);
			rotateLeftUp();
			moveAPiece(FRONT_FACE, quad2, Rubik.RIGHT, false);
			orientLastCenterFace();
		} else if ((rotation0 == 0 && rotation1 == 1) 
				|| (rotation0 == 1 && rotation1 == 2)
				|| (rotation0 == 2 && rotation1 == 3)
				|| (rotation0 == 3 && rotation1 == 0)) {
			System.out.println("unknown case "
				+ rotation0 + " " + rotation1);
		}
	}
	
	void orientCenters() {
		for (int face = 0; face < Rubik.MAX_FACES; face++) {
			rotateRight();
			if (face % 2 == 1)
				rotateCw();
			else
				rotateCcw();
			if (face < 5)
				orientCentersFace(FRONT_FACE, LEFT_FACE, offsetCenters[face]);
		}
		orientLastCenterFace();
	}
	
	/* This procedure coordinates the solution process. */
	void solvePieces() {
		if (!rubik.checkSolved(rubik.orientChoice)) {
			solveTop();
			alignCorners();
			solveBottom();
			if ((rubik.sizeX & 1) == 1) {
				solveBottomEdges();
				alignEdges();
				if (rubik.orientChoice)
					solveCenters();
			} else {
				solve_2_2();
				if (rubik.sizeX > 2) {
					/* 4x4x4 */
					rotateDown();
					rotateLeft();
					solve42U();
					solve2PairEdge4();
					solveLast8Edges();
					solveCenter4();
					if (rubik.orientChoice)
						orientCenters();
				}
			}
		}
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_COMPUTED, rubik.dim);
	}

	private class RubikFE {
		int face, edge;
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solvePieces();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
